#!/usr/bin/env python
'''
WARNING WARNING WARNING

This script changes the input variables of particular tasks and resets certain tasks
so that it's possible to fix errors before starting a new scheduler.

This is an advanced script and you are supposed to be familiar with the Abipy API.
Very bad things will happen if you don't use this script properly.

DO NOT RUN THIS SCRIPT IF THE SCHEDULER IS STILL RUNNING YOUR FLOW.
YOU HAVE BEEN WARNED!
'''

import sys
import argparse

import abipy.flowtk as flowtk

def main():

    # Build the main parser.
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('flowdir', help="File or directory containing the ABINIT flow")
    parser.add_argument('--remove-lock', default=False, action="store_true",
        help="Remove the lock on the pickle file used to save the status of the flow.")
    parser.add_argument('--apply', default=False, action="store_true", help="Apply changes to the flow.")

    options = parser.parse_args()

    # Read the flow from the pickle file
    flow = flowtk.Flow.pickle_load(options.flowdir, remove_lock=options.remove_lock)

    # List with the id of the tasks that will be changed.
    nids = []

    for task in flow.iflat_tasks():

        # Select tasks according to class and status
        # Clearly, you will need to customize this part.
        if task.__class__.__name__ == "PhononTask" and task.status in (task.S_UNCONVERGED, task.S_ERROR):

            nids.append(task.node_id)

            # Here we operate on the AbinitInput of the task.
            # In this particular case, we want to use toldfe instead of tolvrs.
            # Again, you will need to customize this part.

            # Remove all the tolerance variables present in the input.
            task.input.pop_tolerances()

            # Set new stopping criterion
            task.input["toldfe"] = 1e-8
            task.input["nstep"] = 1000
            task.input["prtwf"] = 0

    if options.apply:
        print(f"Resetting {len(nids)} tasks modified by the user.")

        if nids:
            for task in flow.iflat_tasks(nids=nids):
                task.reset()

        print("Writing new pickle file.")

        flow.build_and_pickle_dump()
    else:
        print(f"Dry run mode. {len(nids)} tasks have been modified in memory.")

        if nids:
            for task in flow.iflat_tasks(nids=nids):
                print(task)

        print("Use --apply to reset these tasks and update the pickle file.")
        print("Then restart the scheduler with `nohup abirun.py FLOWDIR scheduler ...`")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Integration tests for phonon flows."""
from __future__ import annotations

import os
import numpy as np
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk

from abipy.core.testing import has_matplotlib

import logging
logger = logging.getLogger(__name__)


def scf_ph_inputs(tvars=None):
    """
    This function constructs the input files for the phonon calculation:
    GS input + the input files for the phonon calculation.
    """
    # Crystalline AlAs: computation of the second derivative of the total energy
    structure = abidata.structure_from_ucell("AlAs")

    # List of q-points for the phonon calculation (4,4,4) mesh.
    qpoints = [
             0.00000000E+00,  0.00000000E+00,  0.00000000E+00,
             2.50000000E-01,  0.00000000E+00,  0.00000000E+00,
             5.00000000E-01,  0.00000000E+00,  0.00000000E+00,
             2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
             5.00000000E-01,  2.50000000E-01,  0.00000000E+00,
            -2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
             5.00000000E-01,  5.00000000E-01,  0.00000000E+00,
            -2.50000000E-01,  5.00000000E-01,  2.50000000E-01,
            ]

    qpoints = np.reshape(qpoints, (-1, 3))

    # Global variables used both for the GS and the DFPT run.
    global_vars = dict(nband=4,
                       ecut=3.0,
                       ngkpt=[4, 4, 4],
                       shiftk=[0, 0, 0],
                       tolvrs=1.0e-6,
                       paral_kgb=0 if tvars is None else tvars.paral_kgb,
                    )

    multi = abilab.MultiDataset(structure=structure, pseudos=abidata.pseudos("13al.981214.fhi", "33as.pspnc"),
                               ndtset=1 + len(qpoints))

    multi.set_vars(global_vars)

    for i, qpt in enumerate(qpoints):
        # Response-function calculation for phonons.
        multi[i+1].set_vars(
            nstep=20,
            rfphon=1,        # Will consider phonon-type perturbation
            nqpt=1,          # One wavevector is to be considered
            qpt=qpt,         # This wavevector is q=0 (Gamma)
            kptopt=3,
            )

            #rfatpol   1 1   # Only the first atom is displaced
            #rfdir   1 0 0   # Along the first reduced coordinate axis
            #kptopt   2      # Automatic generation of k points, taking

    # Split input into gs_inp and ph_inputs
    return multi.split_datasets()


def itest_phonon_flow(fwp, tvars):
    """
    Create a flow for phonon calculations:

        1) One work for the GS run.

        2) nqpt works for phonon calculations. Each work contains
           nirred tasks where nirred is the number of irreducible phonon perturbations
           for that particular q-point.
    """
    all_inps = scf_ph_inputs(tvars)
    scf_input, ph_inputs = all_inps[0], all_inps[1:]

    ph_ngqpt = (2, 2, 2)
    flow = flowtk.PhononFlow.from_scf_input(fwp.workdir, scf_input, ph_ngqpt=ph_ngqpt, with_becs=False)

    scheduler = flow.make_scheduler()
    assert scheduler.start() == 0
    flow.check_status(show=True)
    assert all(work.finalized for work in flow)
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()

    with flow.open_final_ddb() as ddb:
        ddb_path = ddb.filepath
        ddb.to_string(verbose=2)
        assert len(ddb.structure) == 2
        #assert ddb.qpoints.frac_coords

    # Test PhononTask inspect method
    ph_task = flow[1][0]
    if has_matplotlib():
        assert ph_task.inspect(show=False)
    # Test get_results
    #ph_task.get_results()

    # Build new work with Anaddb tasks.
    # Construct a manager with mpi_procs==1 since anaddb do not support mpi_procs > 1 (except in elphon)
    shell_manager = fwp.manager.to_shell_manager(mpi_procs=1)
    awork = flowtk.Work(manager=shell_manager)

    # Phonons bands and DOS with gaussian method
    anaddb_input = abilab.AnaddbInput.phbands_and_dos(
        scf_input.structure, ngqpt=ph_ngqpt, ndivsm=5, nqsmall=10, dos_method="gaussian: 0.001 eV")

    atask = flowtk.AnaddbTask(anaddb_input, ddb_node=ddb_path, manager=shell_manager)
    awork.register(atask)

    # Phonons bands and DOS with tetrahedron method
    anaddb_input = abilab.AnaddbInput.phbands_and_dos(
        scf_input.structure, ngqpt=ph_ngqpt, ndivsm=5, nqsmall=10, dos_method="tetra")

    atask = flowtk.AnaddbTask(anaddb_input, ddb_node=ddb_path, manager=shell_manager)
    awork.register(atask)

    flow.register_work(awork)
    flow.allocate()
    flow.build()

    for i, atask in enumerate(awork):
        atask.history.info("About to run anaddb task: %d", i)
        atask.start_and_wait()
        assert atask.status == atask.S_DONE
        atask.check_status()
        assert atask.status == atask.S_OK

        # These output files should be produced in the task workdir.
        # Actually they should be in the outdir but anaddb uses different conventions.
        assert len(atask.outdir.list_filepaths(wildcard="*PHBST.nc")) == 1
        assert len(atask.outdir.list_filepaths(wildcard="*PHDOS.nc")) == 1


def itest_phonon_restart(fwp):
    """Test the restart of phonon calculations with the scheduler."""
    # Crystalline AlAs: computation of the second derivative of the total energy
    structure = abidata.structure_from_ucell("AlAs")

    # List of q-points for the phonon calculation (4,4,4) mesh.
    qpoints = np.reshape([
        0.00000000E+00,  0.00000000E+00,  0.00000000E+00,
        2.50000000E-01,  0.00000000E+00,  0.00000000E+00,
        #5.00000000E-01,  0.00000000E+00,  0.00000000E+00,  # XXX Uncomment this line to test restart from 1DEN
                                                            # Too long --> disabled
    ], (-1, 3))

    # Global variables used both for the GS and the DFPT run.
    global_vars = dict(
        nband=4,
        ecut=3.0,
        ngkpt=[4, 4, 4],
        shiftk=[0, 0, 0],
        tolvrs=1.0e-5,
    )

    multi = abilab.MultiDataset(structure=structure,
                                pseudos=abidata.pseudos("13al.981214.fhi", "33as.pspnc"),
                                ndtset=1 + len(qpoints))

    multi.set_vars(global_vars)

    for i, qpt in enumerate(qpoints):
        # Response-function calculation for phonons.
        multi[i+1].set_vars(
            rfphon=1,        # Will consider phonon-type perturbation.
            nqpt=1,          # One wavevector is to be considered.
            qpt=qpt,         # q-wavevector.
            kptopt=3,
            nstep=5,         # This is to trigger the phonon restart.
        )
        #rfatpol   1 1   # Only the first atom is displaced
        #rfdir   1 0 0   # Along the first reduced coordinate axis
        #kptopt   2      # Automatic generation of k points, taking

        # i == 0 --> restart from WFK
        if i == 1: multi[i+1].set_vars(prtwf=-1, nstep=5)  # Restart with WFK and smart-io.
        if i == 2: multi[i+1].set_vars(prtwf=0, nstep=8)   # Restart from 1DEN. Too long --> disabled.

    all_inps = multi.split_datasets()
    scf_input, ph_inputs = all_inps[0], all_inps[1:]

    flow = phonon_flow(fwp.workdir, scf_input, ph_inputs, manager=fwp.manager)
    flow.set_garbage_collector()

    for task in flow.iflat_tasks():
        task.manager.qadapter.max_num_launches = 20

    assert flow.make_scheduler().start() == 0

    flow.check_status(show=True, verbose=1)
    assert all(work.finalized for work in flow)
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()

    assert sum(task.num_restarts for task in flow.iflat_tasks()) > 0


def phonon_flow(workdir, scf_input, ph_inputs, with_nscf=False, with_ddk=False, with_dde=False,
                manager=None, flow_class=flowtk.PhononFlow, allocate=True):
    """
    Build a :class:`PhononFlow` for phonon calculations.
    Args:
        workdir: Working directory.
        scf_input: Input for the GS SCF run.
        ph_inputs: List of Inputs for the phonon runs.
        with_nscf: add an nscf task in front of al phonon tasks to make sure the q point is covered
        with_ddk: add the ddk step
        with_dde: add the dde step it the dde is set ddk is switched on automatically
        manager: :class:`TaskManager` used to submit the jobs
                 Initialized from manager.yml if manager is None.
        flow_class: Flow class
    Returns:
        :class:`Flow` object
    """
    if with_dde:
        with_ddk = True

    natom = len(scf_input.structure)

    # Create the container that will manage the different works.
    flow = flow_class(workdir, manager=manager)

    # Register the first work (GS calculation)
    # register_task creates a work for the task, registers it to the flow and returns the work
    # the 0the element of the work is the task
    scf_task = flow.register_task(scf_input, task_class=flowtk.ScfTask)[0]

    # Build a temporary work with a shell manager just to run
    # ABINIT to get the list of irreducible pertubations for this q-point.
    shell_manager = flow.manager.to_shell_manager(mpi_procs=1)

    if with_ddk:
        logger.info('add ddk')
        # TODO
        # MG Warning: be careful here because one should use tolde or tolwfr (tolvrs shall not be used!)
        ddk_input = ph_inputs[0].deepcopy()
        ddk_input.set_vars(qpt=[0, 0, 0], rfddk=1, rfelfd=2, rfdir=[1, 1, 1])
        ddk_task = flow.register_task(ddk_input, deps={scf_task: 'WFK'}, task_class=flowtk.DdkTask)[0]

    if with_dde:
        logger.info('add dde')
        dde_input = ph_inputs[0].deepcopy()
        dde_input.set_vars(qpt=[0, 0, 0], rfddk=1, rfelfd=2)
        dde_input_idir = dde_input.deepcopy()
        dde_input_idir.set_vars(rfdir=[1, 1, 1])
        dde_task = flow.register_task(dde_input, deps={scf_task: 'WFK', ddk_task: 'DDK'},
                                      task_class=flowtk.DdeTask)[0]

    if not isinstance(ph_inputs, (list, tuple)):
        ph_inputs = [ph_inputs]

    for i, ph_input in enumerate(ph_inputs):
        fake_input = ph_input.deepcopy()

        # Run abinit on the front-end to get the list of irreducible pertubations.
        tmp_dir = os.path.join(workdir, "__ph_run" + str(i) + "__")
        w = flowtk.PhononWork(workdir=tmp_dir, manager=shell_manager)
        fake_task = w.register(fake_input)

        # Use the magic value paral_rf = -1 to get the list of irreducible perturbations for this q-point.
        abivars = dict(
            paral_rf=-1,
            rfatpol=[1, natom],  # Set of atoms to displace.
            rfdir=[1, 1, 1],     # Along this set of reduced coordinate axis.
        )

        fake_task.set_vars(abivars)
        w.allocate()
        w.start(wait=True)

        # Parse the file to get the perturbations.
        try:
            irred_perts = flowtk.yaml_read_irred_perts(fake_task.log_file.path)
        except Exception:
            print("Error in %s" % fake_task.log_file.path)
            raise

        logger.info(irred_perts)
        w.rmtree()

        # Now we can build the final list of works:
        # One work per q-point, each work computes all
        # the irreducible perturbations for a singe q-point.

        work_qpt = flowtk.PhononWork()

        if with_nscf:
            # MG: Warning this code assume 0 is Gamma!
            import copy
            nscf_input = copy.deepcopy(scf_input)
            nscf_input.set_vars(kptopt=3, iscf=-3, qpt=irred_perts[0]['qpt'], nqpt=1)
            nscf_task = work_qpt.register_nscf_task(nscf_input, deps={scf_task: "DEN"})
            deps = {nscf_task: "WFQ", scf_task: "WFK"}
        else:
            deps = {scf_task: "WFK"}

        if with_ddk:
            deps[ddk_task] = 'DDK'

        logger.info(irred_perts[0]['qpt'])

        for irred_pert in irred_perts:
            #print(irred_pert)
            new_input = ph_input.deepcopy()

            #rfatpol   1 1   # Only the first atom is displaced
            #rfdir   1 0 0   # Along the first reduced coordinate axis
            qpt = irred_pert["qpt"]
            idir = irred_pert["idir"]
            ipert = irred_pert["ipert"]

            # TODO this will work for phonons, but not for the other types of perturbations.
            rfdir = 3 * [0]
            rfdir[idir - 1] = 1
            rfatpol = [ipert, ipert]

            new_input.set_vars(
                #rfpert=1,
                qpt=qpt,
                rfdir=rfdir,
                rfatpol=rfatpol,
            )

            if with_ddk:
                new_input.set_vars(rfelfd=3)

            work_qpt.register_phonon_task(new_input, deps=deps)

        flow.register_work(work_qpt)

    if allocate: flow.allocate()

    return flow



# coding: utf-8
"""Mixin classes"""
from __future__ import annotations

import abc
import os
import collections
import tempfile
import pickle
import numpy as np
import pandas as pd

from time import ctime
from shutil import which
from monty.termcolor import cprint
from monty.string import list_strings
from monty.collections import dict2namedtuple
from monty.functools import lazy_property
from abipy.tools.typing import Figure


__all__ = [
    "AbinitNcFile",
    "Has_Structure",
    "Has_ElectronBands",
    "Has_PhononBands",
    "NotebookWriter",
    "Has_Header",
    "SlotPickleMixin"
]


class BaseFile(metaclass=abc.ABCMeta):
    """
    Abstract base class defining the methods that must be implemented
    by the concrete classes representing the different files produced by ABINIT.
    """
    def __init__(self, filepath: str):
        filepath = str(filepath)
        self._filepath = os.path.abspath(filepath)

        # Save stat values
        stat = os.stat(filepath)
        self._last_atime = stat.st_atime
        self._last_mtime = stat.st_mtime
        self._last_ctime = stat.st_ctime

    def __repr__(self) -> str:
        return "<%s, %s>" % (self.__class__.__name__, self.relpath)

    @classmethod
    def from_file(cls, filepath: str):
        """Initialize the object from a string."""
        if isinstance(filepath, cls): return filepath
        return cls(filepath)

    @property
    def filepath(self) -> str:
        """Absolute path of the file."""
        return self._filepath

    @property
    def relpath(self) -> str:
        """Relative path."""
        try:
            return os.path.relpath(self.filepath)
        except OSError:
            # current working directory may not be defined!
            return self.filepath

    @property
    def basename(self) -> str:
        """Basename of the file."""
        return os.path.basename(self.filepath)

    @property
    def filetype(self) -> str:
        """String defining the filetype."""
        return self.__class__.__name__

    def filestat(self, as_string: bool = False) -> dict:
        """
        Dictionary with file metadata, if ``as_string`` is True, a string is returned.
        """
        d = get_filestat(self.filepath)
        if not as_string: return d
        return "\n".join("%s: %s" % (k, v) for k, v in d.items())

    @abc.abstractmethod
    def close(self) -> None:
        """Close the file."""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Activated at the end of the with statement. It automatically closes the file."""
        self.close()

    def remove(self) -> None:
        """Close the file handle, remove the file from disk."""
        self.close()
        try:
            os.remove(self.filepath)
        except FileNotFoundError:
            pass


class TextFile(BaseFile):

    def __enter__(self):
        # Open the file
        self._file
        return self

    def __iter__(self):
        return iter(self._file)

    @lazy_property
    def _file(self):
        """File object open in read-only mode."""
        return open(self.filepath, mode="rt")

    def close(self) -> None:
        """Close the file."""
        try:
            self._file.close()
        except Exception:
            pass

    def seek(self, offset, whence: int = 0) -> None:
        """Set the file's current position, like stdio's fseek()."""
        self._file.seek(offset, whence)

    def get_panel(self, **kwargs):
        import panel as pn
        import panel.widgets as pnw

        root, ext = os.path.splitext(self.basename)
        text = open(self.filepath, "rt").read()
        if not text: text = "This file is empty!"

        # Use Markdown for selected extensions else Ace editor.
        if ext and len(ext) > 1: ext = ext[1:]
        ext2format = dict(sh="shell", py="python", stdin="shell", stdout="shell", stderr="shell")

        if ext in ext2format:
            fmt = ext2format[ext]
            obj = pn.pane.Markdown(f"```{fmt}\n{text}\n```", sizing_mode="stretch_both")
        else:
            obj = pnw.Ace(value=text, language='text', readonly=True,
                          sizing_mode='stretch_width', height=1200)

        return pn.Column(f"## File: {self.filepath}",
                         obj,
                         pn.layout.Divider(),
                         sizing_mode="stretch_width")


class JsonFile(TextFile):
    """
    A TextFile containing JSON data.
    Provides get_panel method so that we can visualize the file with `abiopen.py FILE --panel`
    """

    def get_panel(self, **kwargs):
        import json
        from abipy.panels.viewers import JSONViewer
        with self:
            return JSONViewer(json.load(self._file))


class AbinitNcFile(BaseFile):
    """
    Abstract class representing a Netcdf file with data saved
    according to the ETSF-IO specifications (when available).
    An AbinitNcFile has a netcdf reader to read data from file and build objects.
    """

    @classmethod
    def from_binary_string(cls, bstring) -> AbinitNcFile:
        """
        Build object from a binary string with the netcdf data.
        Useful for implementing GUIs in which widgets returns binary data.
        """
        workdir = tempfile.mkdtemp()
        fd, tmp_path = tempfile.mkstemp(suffix=".nc")
        with open(tmp_path, "wb") as fh:
            fh.write(bstring)
            return cls.from_file(tmp_path)

    def ncdump(self, *nc_args, **nc_kwargs) -> str:
        """Returns a string with the output of ncdump."""
        return NcDumper(*nc_args, **nc_kwargs).dump(self.filepath)

    @lazy_property
    def abinit_version(self) -> str:
        """String with the abinit version: three digits separated by comma."""
        return self.reader.rootgrp.getncattr("abinit_version")

    @abc.abstractproperty
    def params(self) -> dict:
        """
        dict with the convergence parameters used to construct |pandas-DataFrames|.
        """

    def get_dims_dataframe(self, as_dict=False, path="/") -> pd.DataFrame:
        """
        Return: |pandas-Dataframe| with the dimensions defined in the `path` group.
            or dict if as_dict is True.
        """
        grp = self.reader.rootgrp if path == "/" else self.path2group[path]
        d = {k: len(v) for k, v in grp.dimensions.items()}

        if as_dict: return d

        # Since this is a Series but we want a dataframe to facilitate interoperability.
        # we have to call init with additional kwargs.
        return pd.DataFrame.from_dict(d, orient='index', columns=['value'])

    def get_input_string(self) -> str:
        """
        Read and return input string stored in the netcdf.
        Only nc files generared by Abinit9 have this variable.
        """
        if "input_string" in self.reader.rootgrp.variables:
            return self.reader.read_string("input_string")
        else:
            return "Nc file does not contain `input_string`"

    def get_ncfile_view(self, **kwargs):
        """
        Return panel Parameterized object with widgets to visualize
        netcdf dimensions and variables.
        """
        from abipy.panels.core import NcFileViewer
        return NcFileViewer(self).get_ncfile_view(**kwargs)


class AbinitFortranFile(BaseFile):
    """
    Abstract class representing a Fortran file containing output data from abinit.
    """
    def close(self) -> None:
        """nop, just to fulfill the abstract interface."""


class CubeFile(BaseFile):
    """
    .. attribute:: structure

        |Structure| object

    .. attribute:: mesh

        |Mesh3d| object with information on the uniform 3d mesh.

    .. attribute:: data

        |numpy-array| of shape [nx, ny, nz] with numerical values on the real-space mesh.
    """
    def __init__(self, filepath: str):
        from abipy.iotools.cube import cube_read_structure_mesh_data
        super().__init__(filepath)
        self.structure, self.mesh, self.data = cube_read_structure_mesh_data(self.filepath)

    def close(self) -> None:
        """nop, just to fulfill the abstract interface."""

    #@classmethod
    #def write_structure_mesh_data(cls, path, structure, mesh, data):
    #    with open(path, "wt") as fh:
    #        cube_write_structure_mesh(fh, structure, mesh)
    #        cube_write_data(fh, data, mesh):


class Has_Structure(metaclass=abc.ABCMeta):
    """
    Mixin class for |AbinitNcFile| containing crystallographic data.
    """

    @abc.abstractproperty
    def structure(self):
        """Returns the |Structure| object."""

    def plot_bz(self, **kwargs) -> Figure:
        """
        Gives the plot (as a matplotlib object) of the symmetry line path in the Brillouin Zone.
        """
        return self.structure.plot_bz(**kwargs)

    # To maintain backward compatbility
    show_bz = plot_bz

    def export_structure(self, filepath: str):
        """
        Export the structure on file.

        returns: |Visualizer| instance.
        """
        return self.structure.export(filepath)

    def visualize_structure_with(self, appname: str):
        """
        Visualize the crystalline structure with the specified visualizer.

        See |Visualizer| for the list of applications and formats supported.
        """
        from abipy.iotools.visualizer import Visualizer
        visu = Visualizer.from_name(appname)

        for ext in visu.supported_extensions():
            ext = "." + ext
            try:
                return self.export_structure(ext)
            except visu.Error:
                pass
        else:
            raise visu.Error("Don't know how to export data for appname %s" % appname)

    def _get_atomview(self, view, select_symbols=None, verbose=0):
        """
        Helper function used to select (inequivalent||all) atoms depending on view.
        Uses spglib to find inequivalent sites.

        Args:
            view: "inequivalent" to show only inequivalent atoms. "all" for all sites.
            select_symbols: String or list of strings with chemical symbols.
                Used to select only atoms of this type.

        Return named tuple with:

                * iatom_list: list of site index.
                * wyckoffs: Wyckoff letters
                * site_labels: Labels for each site in `iatom_list` e.g Si2a
        """
        natom = len(self.structure)
        if natom == 1: verbose = False
        if verbose:
            print("Calling spglib to find inequivalent sites. Magnetic symmetries (if any) are not taken into account.")

        ea = self.structure.spget_equivalent_atoms(printout=verbose > 0)

        # Define iatom_list depending on view
        if view == "all":
            iatom_list = np.arange(natom)
        elif view == "inequivalent":
            iatom_list = ea.irred_pos
        else:
            raise ValueError("Wrong value for view: %s" % str(view))

        # Filter by element symbol.
        if select_symbols is not None:
            select_symbols = set(list_strings(select_symbols))
            iatom_list = [i for i in iatom_list if self.structure[i].specie.symbol in select_symbols]
            iatom_list = np.array(iatom_list, dtype=int)

        # Slice full arrays.
        wyckoffs = ea.wyckoffs[iatom_list]
        wyck_labels = ea.wyck_labels[iatom_list]
        site_labels = ea.site_labels[iatom_list]

        return dict2namedtuple(iatom_list=iatom_list, wyckoffs=wyckoffs, wyck_labels=wyck_labels, site_labels=site_labels)

    def yield_structure_figs(self, **kwargs):
        """*Generates* a predefined list of matplotlib figures with minimal input from the user."""
        yield self.structure.plot(show=False)

    def yield_structure_plotly_figs(self, **kwargs):
        """*Generates* a predefined list of plotly figures with minimal input from the user."""
        yield self.structure.plotly(show=False)


class Has_ElectronBands(metaclass=abc.ABCMeta):
    """Mixin class for |AbinitNcFile| containing electron data."""

    @abc.abstractproperty
    def ebands(self):
        """Returns the |ElectronBands| object."""

    @property
    def nsppol(self) -> int:
        """Number of spin polarizations"""
        return self.ebands.nsppol

    @property
    def nspinor(self) -> int:
        """Number of spinors"""
        return self.ebands.nspinor

    @property
    def nspden(self) -> int:
        """Number of indepedendent spin-density components."""
        return self.ebands.nspden

    @property
    def mband(self) -> int:
        """Maximum number of bands."""
        return self.ebands.mband

    @property
    def nband(self) -> int:
        """Maximum number of bands."""
        return self.ebands.nband

    @property
    def nelect(self) -> float:
        """Number of electrons per unit cell"""
        return self.ebands.nelect

    @property
    def nkpt(self) -> int:
        """Number of k-points."""
        return self.ebands.nkpt

    @property
    def kpoints(self):
        """Iterable with the Kpoints."""
        return self.ebands.kpoints

    @lazy_property
    def tsmear(self):
        return self.ebands.smearing.tsmear_ev.to("Ha")

    def get_ebands_params(self) -> dict:
        """:class:`OrderedDict` with the convergence parameters."""
        return collections.OrderedDict([
            ("nsppol", self.nsppol),
            ("nspinor", self.nspinor),
            ("nspden", self.nspden),
            ("nband", self.nband),
            ("nkpt", self.nkpt),
        ])

    def plot_ebands(self, **kwargs) -> Figure:
        """Plot the electron energy bands. See the :func:`ElectronBands.plot` for the signature."""
        return self.ebands.plot(**kwargs)

    def plot_ebands_with_edos(self, edos, **kwargs) -> Figure:
        """Plot the electron energy bands with DOS. See the :func:`ElectronBands.plot_with_edos` for the signature."""
        return self.ebands.plot_with_edos(edos, **kwargs)

    def get_edos(self, **kwargs):
        """Compute the electronic DOS on a linear mesh. Wraps ebands.get_edos."""
        return self.ebands.get_edos(**kwargs)

    def yield_ebands_figs(self, **kwargs):
        """*Generates* a predefined list of matplotlib figures with minimal input from the user."""
        with_gaps = not self.ebands.is_metal
        if self.ebands.kpoints.is_path:
            yield self.ebands.plot(with_gaps=with_gaps, show=False)
            if len(self.ebands.kpoints) > 1:
                yield self.ebands.kpoints.plot(show=False)
        else:
            edos = self.ebands.get_edos()
            yield self.ebands.plot_with_edos(edos, with_gaps=with_gaps, show=False)
            yield edos.plot(show=False)

    def yield_ebands_plotly_figs(self, **kwargs):
        """*Generates* a predefined list of plotly figures with minimal input from the user."""
        with_gaps = not self.ebands.is_metal

        if self.ebands.kpoints.is_path:
            yield self.ebands.plotly(with_gaps=with_gaps, show=False)
            if len(self.ebands.kpoints) > 1:
                yield self.ebands.kpoints.plotly(show=False)
        else:
            edos = self.ebands.get_edos()
            yield self.ebands.plotly_with_edos(edos, with_gaps=with_gaps, show=False)
            yield edos.plotly(show=False)

    def expose_ebands(self, slide_mode=False, slide_timeout=None, expose_web=False, **kwargs):
        """
        Shows a predefined list of matplotlib figures for electron bands with minimal input from the user.
        """
        from abipy.tools.plotting import MplExposer, PanelExposer

        if expose_web:
            e = PanelExposer(title=f"e-Bands of {self.structure.formula}")
        else:
            e = MplExposer(slide_mode=slide_mode, slide_timeout=slide_mode, verbose=1)

        with e:
            e(self.yield_ebands_figs(**kwargs))

    #def plotly_expose_ebands(self, **kwargs):
    #    """
    #    This function *generates* a predefined list of plotly figures with minimal input from the user.
    #    """
    #    chart_studio = kwargs.pop("chart_studio", None)
    #    verbose = kwargs.pop("verbose", 0)
    #    kwargs.update(dict(
    #        renderer="chart_studio" if chart_studio else None,
    #        title=f"Band structure of {self.ebands.structure.formula}",
    #        with_gaps = not self.ebands.has_metallic_scheme,
    #    ))
    #    self.ebands.plotly(**kwargs)


class Has_PhononBands(metaclass=abc.ABCMeta):
    """
    Mixin class for |AbinitNcFile| containing phonon data.
    """

    @abc.abstractproperty
    def phbands(self):
        """Returns the |PhononBands| object."""

    def get_phbands_params(self) -> dict:
        """:class:`OrderedDict` with the convergence parameters."""
        return collections.OrderedDict([
            ("nqpt", len(self.phbands.qpoints)),
        ])

    def plot_phbands(self, **kwargs) -> Figure:
        """
        Plot the electron energy bands. See the :func:`PhononBands.plot` for the signature.""
        """
        return self.phbands.plot(**kwargs)

    #def plot_phbands_with_phdos(self, phdos, **kwargs):
    #    return self.phbands.plot_with_phdos(phdos, **kwargs)

    def yield_phbands_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        units = kwargs.get("units", "mev")
        yield self.phbands.qpoints.plot(show=False)
        yield self.phbands.plot(units=units, show=False)
        yield self.phbands.plot_colored_matched(units=units, show=False)

    def yield_phbands_plotly_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of plotly figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        units = kwargs.get("units", "mev")
        yield self.phbands.qpoints.plotly(show=False)
        yield self.phbands.plotly(units=units, show=False)
        yield self.phbands.plot_colored_matched(units=units, show=False)

    def expose_phbands(self, slide_mode=False, slide_timeout=None, **kwargs):
        """
        Shows a predefined list of matplotlib figures for phonon bands with minimal input from the user.
        """
        from abipy.tools.plotting import MplExposer
        with MplExposer(slide_mode=slide_mode, slide_timeout=slide_mode, verbose=1) as e:
            e(self.yield_phbands_figs(**kwargs))


class NcDumper(object):
    """
    Wrapper object for the ncdump tool.
    """

    def __init__(self, *nc_args, **nc_kwargs):
        """
        Args:
            nc_args: Arguments passed to ncdump.
            nc_kwargs: Keyword arguments passed to ncdump
        """
        self.nc_args = nc_args
        self.nc_kwargs = nc_kwargs
        self.ncdump = which("ncdump")

    def dump(self, filepath: str) -> str:
        """Returns a string with the output of ncdump."""
        if self.ncdump is None:
            return "Cannot find ncdump tool in $PATH"
        else:
            from subprocess import check_output
            return check_output(["ncdump", filepath])


_ABBREVS = [
    (1 << 50, 'Pb'),
    (1 << 40, 'Tb'),
    (1 << 30, 'Gb'),
    (1 << 20, 'Mb'),
    (1 << 10, 'kb'),
    (1, 'b'),
]


def size2str(size):
    """Convert size to string with units."""
    for factor, suffix in _ABBREVS:
        if size > factor:
            break
    return "%.2f " % (size / factor) + suffix


def get_filestat(filepath: str) -> dict:
    stat = os.stat(filepath)
    return collections.OrderedDict([
        ("Name", os.path.basename(filepath)),
        ("Directory", os.path.dirname(filepath)),
        ("Size", size2str(stat.st_size)),
        ("Access Time", ctime(stat.st_atime)),
        ("Modification Time", ctime(stat.st_mtime)),
        ("Change Time", ctime(stat.st_ctime)),
    ])


class HasNotebookTools:

    def has_panel(self):
        """
        Return panel module (that evaluates to True) if panel is installed else False.
        """
        try:
            import panel as pn
            return pn
        except ImportError:
            return False

    def make_and_open_notebook(self, nbpath=None, foreground=False,
                               classic_notebook=False, no_browser=False) -> int:  # pragma: no cover
        """
        Generate a jupyter_ notebook and open it in the browser.

        Args:
            nbpath: If nbpath is None, a temporay file is created.
                foreground: By default, jupyter is executed in background and stdout, stderr are redirected.
                to devnull. Use foreground to run the process in foreground
            classic_notebook: True to use the classic notebook instead of jupyter-lab (default)
            no_browser: Start the jupyter server to serve the notebook but don't open the notebook in the browser.
                        Use this option to connect remotely from localhost to the machine running the kernel

        Return: system exit code.

        Raise: `RuntimeError` if jupyter executable is not in $PATH
        """
        nbpath = self.write_notebook(nbpath=nbpath)

        if not classic_notebook:
            # Use jupyter-lab.
            app_path = which("jupyter-lab")
            if app_path is None:
                raise RuntimeError("""
Cannot find jupyter-lab application in $PATH. Install it with:

    conda install -c conda-forge jupyterlab

or:

    pip install jupyterlab

See also https://jupyterlab.readthedocs.io/
""")

        else:
            # Use classic notebook
            app_path = which("jupyter")
            if app_path is None:
                raise RuntimeError("""
Cannot find jupyter application in $PATH. Install it with:

    conda install -c conda-forge jupyter

or:

    pip install jupyterlab

See also https://jupyter.readthedocs.io/en/latest/install.html
""")
            app_path = app_path + " notebook "

        if not no_browser:
            if foreground:
                return os.system("%s %s" % (app_path, nbpath))
            else:
                fd, tmpname = tempfile.mkstemp(text=True)
                print(tmpname)
                cmd = "%s %s" % (app_path, nbpath)
                print("Executing:", cmd, "\nstdout and stderr redirected to %s" % tmpname)
                import subprocess
                process = subprocess.Popen(cmd.split(), shell=False, stdout=fd, stderr=fd)
                cprint("pid: %s" % str(process.pid), "yellow")
                return 0

        else:
            # Based on https://github.com/arose/nglview/blob/master/nglview/scripts/nglview.py
            notebook_name = os.path.basename(nbpath)
            dirname = os.path.dirname(nbpath)
            print("nbpath:", nbpath)

            import socket
            from abipy.tools.notebooks import find_free_port
            username = os.getlogin()
            hostname = socket.gethostname()
            port = find_free_port()

            client_cmd = "ssh -NL localhost:{port}:localhost:{port} {username}@{hostname}".format(
                username=username, hostname=hostname, port=port)

            print(f"""
Using port: {port}

\033[32m In your local machine, run: \033[0m

                {client_cmd}

\033[32m NOTE: you might want to replace {hostname} by full hostname with domain name \033[0m
\033[32m Then open your web browser, copy and paste the URL: \033[0m

http://localhost:{port}/notebooks/{notebook_name}
""")
            if not classic_notebook:
                cmd = f'{app_path} {notebook_name} --no-browser --port {port} --notebook-dir {dirname}'
            else:
                cmd = f'{app_path} notebook {notebook_name} --no-browser --port {port} --notebook-dir {dirname}'

            print("Executing:", cmd)
            print('NOTE: make sure to open `{}` in your local machine\n'.format(notebook_name))

            return os.system(cmd)

    @staticmethod
    def get_nbformat_nbv() -> tuple:
        """Return (nbformat module, notebook version module)"""
        import nbformat
        nbv = nbformat.v4
        return nbformat, nbv

    def get_nbformat_nbv_nb(self, title=None) -> tuple:
        """
        Return ``nbformat`` module, notebook version module
        and new notebook with title and import section
        """
        nbformat, nbv = self.get_nbformat_nbv()
        nb = nbv.new_notebook()

        if title is not None:
            nb.cells.append(nbv.new_markdown_cell("## %s" % title))

        nb.cells.extend([
            nbv.new_code_cell("""\
import sys, os
import numpy as np

%matplotlib notebook

# Use this magic for jupyterlab.
# For installation instructions, see https://github.com/matplotlib/jupyter-matplotlib
#%matplotlib widget

from IPython.display import display

# This to render pandas DataFrames with https://github.com/quantopian/qgrid
#import qgrid
#qgrid.nbinstall(overwrite=True)  # copies javascript dependencies to your /nbextensions folder

# This to view Mayavi visualizations. See http://docs.enthought.com/mayavi/mayavi/tips.html
#from mayavi import mlab; mlab.init_notebook(backend='x3d', width=None, height=None, local=True)

from abipy import abilab

# Tell AbiPy we are inside a notebook and use seaborn settings for plots.
# See https://seaborn.pydata.org/generated/seaborn.set.html#seaborn.set
abilab.enable_notebook(with_seaborn=True)
""")
        ])

        return nbformat, nbv, nb

    @staticmethod
    def _write_nb_nbpath(nb, nbpath: str) -> str:
        """
        This method must be called at the end of ``write_notebook``.
        nb is the jupyter notebook and nbpath the argument passed to ``write_notebook``.
        """
        import io, os, tempfile
        if nbpath is None:
            _, nbpath = tempfile.mkstemp(prefix="abinb_", suffix='.ipynb', dir=os.getcwd(), text=True)

        # Write notebook
        import nbformat
        with io.open(nbpath, 'wt', encoding="utf8") as fh:
            nbformat.write(nb, fh)
            return nbpath


class NotebookWriter(HasNotebookTools, metaclass=abc.ABCMeta):
    """
    Mixin class for objects that are able to generate jupyter_ notebooks.
    Subclasses must provide a concrete implementation of `write_notebook`.
    """

    @abc.abstractmethod
    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to nbpath. If nbpath is None, a temporay file is created.
        Return path to the notebook. A typical template:

        .. code-block:: python

            # Preable.
            nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

            #####################
            # Put your code here
            nb.cells.extend([
                nbv.new_markdown_cell("# This is a markdown cell"),
                nbv.new_code_cell("a = 1"),
            ])
            #####################

            # Call _write_nb_nbpath
            return self._write_nb_nbpath(nb, nbpath)
        """

    @classmethod
    def pickle_load(cls, filepath: str):
        """
        Loads the object from a pickle file.
        """
        with open(filepath, "rb") as fh:
            new = pickle.load(fh)
            #assert cls is new.__class__
            return new

    def pickle_dump(self, filepath=None) -> str:
        """
        Save the status of the object in pickle format.
        If filepath is None, a temporary file is created.

        Return: The name of the pickle file.
        """
        if filepath is None:
            _, filepath = tempfile.mkstemp(suffix='.pickle')

        with open(filepath, "wb") as fh:
            pickle.dump(self, fh)
            return filepath

    @abc.abstractmethod
    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """

    #@abc.abstractmethod
    #def yield_plotly_figs(self, **kwargs):  # pragma: no cover
    #    """
    #    This function *generates* a predefined list of matplotlib figures with minimal input from the user.
    #    Used in abiview.py to get a quick look at the results.
    #    """

    def _get_panel_and_template(self) -> tuple:
        # Create panel template with matplotlib figures and show them in the browser.
        import panel as pn
        pn.config.sizing_mode = 'stretch_width'
        from abipy.panels.core import get_template_cls_from_name
        cls = get_template_cls_from_name("FastGridTemplate")

        title = self.__class__.__name__
        if hasattr(self, "structure"): title = f"{title} <small>({self.structure.formula})</small>"
        template = cls(
            title=title,
            header_background="#ff8c00 ", # Dark orange
        )

        return pn, template

    def expose(self, slide_mode=False, slide_timeout=None, use_web=False, **kwargs):
        """
        Shows a predefined list of figures with minimal input from the user.
        Relies on the ``yield_fig``s methods implemented by the subclass to generate such figures.

        Args:
            use_web: True to show all figures inside a panel template executed in the local browser.
               False to show figures in different GUIs.
        """
        if not use_web:
            # Produce all matplotlib figures and show them with the X-server.
            from abipy.tools.plotting import MplExposer
            with MplExposer(slide_mode=slide_mode, slide_timeout=slide_mode, verbose=1) as e:
                e(self.yield_figs(**kwargs))

        else:
            # Create panel template with matplotlib figures and show them in the browser.
            pn, template = self._get_panel_and_template()
            pn.config.sizing_mode = 'stretch_width'
            from abipy.panels.core import mpl, dfc
            for i, fig in enumerate(self.yield_figs()):
                row, col = divmod(i, 2)
                #if isinstance(fig, pd.DataFrame, pd.Series):
                #    p = dfc(fig)
                #elsse
                p = mpl(fig, with_divider=False, dpi=82)
                if hasattr(template.main, "append"):
                    template.main.append(p)
                else:
                    # Assume .main area acts like a GridSpec
                    row_slice = slice(3 * row, 3 * (row + 1))
                    if col == 0: template.main[row_slice, :6] = p
                    if col == 1: template.main[row_slice, 6:] = p

            return template.show()

    def plotly_expose(self, **kwargs):
        """
        This function *generates* a predefined list of plotly figures with minimal input from the user.
        Relies on the yield_plotly_figs method implemented by the subclass in order to generate the figures.
        """
        #print("in plotly expose")
        pn, template = self._get_panel_and_template()
        pn.config.sizing_mode = 'stretch_width'
        from abipy.panels.core import mpl, ply

        # Insert figure in template.main.
        from abipy.tools.plotting import is_mpl_figure, is_plotly_figure
        for i, fig in enumerate(self.yield_plotly_figs()):
            row, col = divmod(i, 2)
            # Handle both matplotlib and plotly figures since we dont' support plotly everywhere.
            if is_plotly_figure(fig):
                p = ply(fig, with_divider=False)
            elif is_mpl_figure(fig):
                p = mpl(fig, with_divider=False)
            else:
                raise TypeError(f"Don't know how to handle type: `{type(fig)}`")

            if hasattr(template.main, "append"):
                template.main.append(p)
            else:
                # Assume main area acts like a panel GridSpec
                row_slice = slice(3 * row, 3 * (row + 1))
                if col == 0: template.main[row_slice, :6] = p
                if col == 1: template.main[row_slice, 6:] = p

        return template.show()


class Has_Header:
    """
    Mixin class for netcdf files containing the Abinit header.
    """

    @lazy_property
    def hdr(self):
        """|AttrDict| with the Abinit header e.g. hdr.ecut."""
        return self.reader.read_abinit_hdr()

    #def compare_hdr(self, other_hdr):


class SlotPickleMixin:
    """
    This mixin makes it possible to pickle/unpickle objects with __slots__
    """

    def __getstate__(self) -> dict:
        return {slot: getattr(self, slot) for slot in self.__slots__ if hasattr(self, slot)}

    def __setstate__(self, state: dict) -> None:
        for slot, value in state.items():
            setattr(self, slot, value)



"""Objects for post-processing Raman results produced by anaddb."""
from __future__ import annotations

import numpy as np
import abipy.core.abinit_units as abu

from collections import namedtuple
from abipy.iotools import ETSF_Reader
from abipy.core.func1d import Function1D
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt
from abipy.tools.typing import Figure

PowderIntensity = namedtuple("PowderIntensity", ("paral", "perp", "tot"))


class Raman:
    """
    An object that allows to obtain the Raman intensities based on the calculated
    susceptibilities. The expressions used to extract the different values of the
    Raman intensities can be found for example in PRB 71, 214307, Physics of the
    Earth and Planetary Interiors 174, 113 or PRB 71, 125107.
    """

    def __init__(self, susceptibility, phfreqs, non_anal_susceptibility=None,
                 non_anal_phfreqs=None, non_anal_directions=None):
        """
        Args:
            susceptibility: a numpy array with shape (n modes, 3 3) containing the
                Raman susceptibilities of transverse zone-center phonon modes.
            phfreqs: a numpy array with the phonon frequencies at Gamma, without non
                analytical contributions.
            non_anal_susceptibility: a numpy array with shape (n directions, n modes, 3 3)
                containing the Raman susceptibilities of zone-center phonons, with non-analyticity
                in  different directions.
            non_anal_phfreqs: a numpy array with the phonon frequencies with shape
                (n directions, n modes), with non analytical contributions along the different
                directions.
            non_anal_directions: a numpy array with shape (n directions, 3) with the directions
                along which the non analytical contribution has been calculated.
        """
        self.susceptibility = susceptibility
        self.phfreqs = phfreqs
        self.non_anal_susceptibility = non_anal_susceptibility
        self.non_anal_phfreqs = non_anal_phfreqs
        self.non_anal_directions = non_anal_directions

    @classmethod
    def from_file(cls, filepath: str) -> Raman:
        """
        Create the object from an anaddb.nc netcdf file.

        Args:
            filepath: path to the netcdf file.

        Returns: An instance of Raman.
        """

        with ETSF_Reader(filepath) as r:
            try:
                susceptibility = r.read_value("raman_sus").T
                phfreqs = r.read_value("gamma_phonon_modes")
                non_anal_susceptibility = r.read_value("non_analytical_raman_sus", default=None)
                if non_anal_susceptibility is not None:
                    non_anal_susceptibility = non_anal_susceptibility.T
                non_anal_phfreqs = r.read_value("non_analytical_phonon_modes", default=None)
                non_anal_directions = r.read_value("non_analytical_directions", default=None)
            except Exception:
                import traceback
                msg = traceback.format_exc()
                msg += ("Error while trying to read Raman from file.\n"
                        "Verify that the required variables are used in anaddb: nlflag\n")
                raise ValueError(msg)

            return cls(susceptibility=susceptibility, phfreqs=phfreqs, non_anal_susceptibility=non_anal_susceptibility,
                       non_anal_phfreqs=non_anal_phfreqs, non_anal_directions=non_anal_directions)

    def get_modes_intensities(self, temp, laser_freq, non_anal_dir=None, relative=False, units="eV",
                              pol_in=None, pol_out=None) -> np.ndarray:
        """
        Calculates the Raman intensities for each mode in arbitrary units. It is possible to use the
        susceptibilities from the transverse modes only or to specify one of the directions with non
        analytical contributions. By default it returns an array with shape (n modes, 3, 3) where each
        component of the 3x3 matrix are those with different incoming (the first index) and outgoing
        (the second index) polarization. These are along the components along the standard axes (e.g.
        "xx" or "yz" components). If pol_in and pol_out an array with shape (n modes) is returned
        corresponding to the selected polarizations.

        Args:
            temp: temperature in K.
            laser_freq: frequency of the incident laser. The units are determined by the "units"
                argument.
            non_anal_dir: index of the direction along which the non analytical contribution
                has been calculated. Corresponds to the indices in the non_anal_directions attribute.
            relative: if True the intensities will be rescaled so that the largest value is 1.
            units: the units in which the input and the output frequencies will be given.
                Possible values in ("eV", "meV", "Ha", "cm-1", "Thz").
            pol_in: the polarization of the incoming photon. If not None can be either either a string
                with one of the cartesian components i.e. "x", "y", "z" or an array with 3 elements
                representing the polarization direction. If not None pol_out can not be None.
            pol_out: the polarization of the outgoing photon. If not None can be either either a string
                with one of the cartesian components i.e. "x", "y", "z" or an array with 3 elements
                representing the polarization direction. If not None pol_in can not be None.

        Returns:
            An array with the Raman intensities. If pol_in==pol_out==None has shape (n modes, 3, 3)
            with all the components. Otherwise an array with size (n modes) with the intensities of
            the selected polarizations.
        """

        if non_anal_dir is None:
            w = self.phfreqs
            sus = self.susceptibility
        else:
            w = self.non_anal_phfreqs[non_anal_dir]
            sus = self.non_anal_susceptibility[non_anal_dir]

        laser_freq = laser_freq / abu.phfactor_ev2units(units)

        c = self._get_prefactor(w=w, temp=temp, laser_freq=laser_freq)
        if pol_in is None and pol_out is None:
            i = c[:, np.newaxis, np.newaxis] * sus**2
            # this will make the indices of the i,j component such that the first
            # will refer to the polarization of the incoming photon and the second
            # to the polarization of the created one.
            i = np.transpose(i, axes=(0, 2, 1))
        else:
            if pol_in is None or pol_out is None:
                raise ValueError("pol_in and pol_out should be either both None or both defined")
            dxyz = {"x": [1, 0, 0], "y": [0, 1, 0], "z": [0, 0, 1]}
            if isinstance(pol_in, str):
                pol_in = dxyz[pol_in.lower()]
            if isinstance(pol_out, str):
                pol_out = dxyz[pol_out.lower()]

            pol_in = np.array(pol_in) / np.linalg.norm(pol_in)
            pol_out = np.array(pol_out) / np.linalg.norm(pol_out)

            i = c * np.einsum("ijk, j, k -> i", sus, pol_out, pol_in) ** 2

        if relative:
            i /= i.max()

        return i

    @staticmethod
    def _get_prefactor(w, temp, laser_freq) -> np.array:
        """
        Helper method to calculate the coefficient for the Raman intensities.

        Args:
            w: the selected frequencies in eV.
            temp: the temperature in K.
            laser_freq: the frequency of the laser in eV

        Returns:
            An array with shape (n modes) with the coefficient for the Raman intensities.
        """

        c = np.zeros_like(w)
        ind = np.where(w > 1e-5)

        bose_factor = 1 / (1 - np.exp(-w[ind] / (abu.kb_eVK * temp)))

        c[ind] = (w[ind] - laser_freq) ** 4 / (2 * w[ind]) * bose_factor

        return c

    def _get_lorentz_freqs_and_factor(self, intensity, non_anal_dir, min_freq, max_freq, 
                                      num, width, units) -> tuple:
        """
        Helper method to get the list of frequencies and the main spread factors to
        calculate the broadened Raman intensities with a Lorentz distribution.

        Args:
            intensity: the Raman intensities at specified frequencies.
            non_anal_dir: ndex of the direction along which the non analytical contribution
                has been calculated. Corresponds to the indices in the non_anal_directions attribute.
            min_freq: minimum frequency considered. If None it will be given by the minimum
                frequency with non zero intensities minus 10 times the width of the distribution.
            max_freq: maximum frequency considered. If None it will be given by the maximum
                frequency with non zero intensities plus 10 times the width of the distribution.
            width: the width of the Lorentz distribution.
            units: the units in which the input and the output frequencies will be given.
                Possible values in ("eV", "meV", "Ha", "cm-1", "Thz")

        Returns:
            Tuple with list of "num" frequencies in eV and factors for the Lorentz broadening
            with shape (n modes, num).
        """

        if non_anal_dir is None:
            w = self.phfreqs
        else:
            w = self.non_anal_phfreqs[non_anal_dir]

        units_factor = abu.phfactor_ev2units(units)

        width = width / units_factor

        if min_freq is None:
            min_ind = np.where(intensity/intensity.max() > 1e-10)[0].min()
            min_freq = w[min_ind] - 10 * width
        else:
            min_freq = min_freq / units_factor

        if max_freq is None:
            max_ind = np.where(intensity/intensity.max() > 1e-10)[0].max()
            max_freq = w[max_ind] + 10 * width
        else:
            max_freq = max_freq / units_factor

        freqs = np.linspace(min_freq, max_freq, num)

        lorentz = width / ((freqs - w.reshape((-1, 1)))**2 + width**2) / np.pi

        return freqs, lorentz

    def get_lorentz_intensity(self, temp, laser_freq, width, non_anal_dir=None, min_freq=None, max_freq=None,
                              num=1000, relative=False, units="eV", pol_in=None, pol_out=None):
        """
        Calculates the broadened Raman intensities in arbitrary units for frequencies in an interval. It is
        possible to use the susceptibilities from the transverse modes only or to specify one of the directions
        with non analytical contributions. By default it returns a 3x3 matrix where each component is a
        Function1D object with the Raman intensities with different incoming (the first index) and outgoing
        (the second index) polarization. These are along the components along the standard axes (e.g. "xx" or
        "yz" components). If pol_in and pol_out a single Function1D is returned corresponding to the selected
        polarizations.

        Args:
            temp: temperature in K.
            laser_freq: frequency of the incident laser. The units are determined the "units"
                argument.
            width: the width of the Lorentz distribution. The units are determined the "units"
                argument.
            non_anal_dir: index of the direction along which the non analytical contribution
                has been calculated. Corresponds to the indices in the non_anal_directions attribute.
            min_freq: minimum frequency considered. If None it will be given by the minimum
                frequency with non zero intensities minus 10 times the width of the distribution.
                If given the units are determined by the "units" argument.
            max_freq: maximum frequency considered. If None it will be given by the maximum
                frequency with non zero intensities plus 10 times the width of the distribution.
                If given the units are determined by the "units" argument.
            num: number of frequencies in the interval (min_freq, max_freq).
            relative: if True the intensities will be rescaled so that the largest value is 1.
            units: the units in which the input and the output frequencies will be given.
                Possible values in ("eV", "meV", "Ha", "cm-1", "Thz")
            pol_in: the polarization of the incoming photon. If not None can be either either a string
                with one of the cartesian components i.e. "x", "y", "z" or an array with 3 elements
                representing the polarization direction. If not None pol_out can not be None.
            pol_out: the polarization of the outgoing photon. If not None can be either either a string
                with one of the cartesian components i.e. "x", "y", "z" or an array with 3 elements
                representing the polarization direction. If not None pol_in can not be None.

        Returns:
            If pol_in==pol_out==None a 3x3 list with a Function1D corresponding to the different
            components of the intensities. Otherwise a single Function1D with the intensities of
            the selected polarizations. Each Function1D has "num" points.
        """

        i = self.get_modes_intensities(temp=temp, laser_freq=laser_freq, non_anal_dir=non_anal_dir,
                                       units=units, pol_in=pol_in, pol_out=pol_out)

        freqs, lorentz = self._get_lorentz_freqs_and_factor(intensity=i, non_anal_dir=non_anal_dir, min_freq=min_freq,
                                                            max_freq=max_freq, num=num, width=width, units=units)

        # convert the frequencies to the desired units for the output
        x = freqs * abu.phfactor_ev2units(units)

        if pol_in is not None and pol_out is not None:
            li = np.dot(i, lorentz)
            if relative:
                li /= li.max()

            return Function1D(x, li)

        else:
            li = np.einsum("ij, ikl -> jkl", lorentz, i)

            li_func = [[None]*3]*3

            for i in range(3):
                for j in range(3):
                    y = li[:, i, j]
                    if relative:
                        y /= y.max()
                    li_func[i][j] = Function1D(x, y)

            return li_func

    def get_powder_intensity(self, temp, laser_freq, non_anal_dir=None, 
                             relative=False, units="eV") -> PowderIntensity:
        """
        Calculates the Raman intensities in arbitrary units for each mode integrated over all possible
        orientation to reproduce the powder measurements. It is possible to use the susceptibilities from
        transverse modes only or to specify one of the directions with non analytical contributions.

        Args:
            temp: temperature in K.
            laser_freq: frequency of the incident laser. The units are determined the "units"
                argument.
            non_anal_dir: index of the direction along which the non analytical contribution
                has been calculated. Corresponds to the indices in the non_anal_directions attribute.
            relative: if True the intensities will be rescaled so that the largest value of the
                total intensity is 1.
            units: the units in which the input and the output frequencies will be given.
                Possible values in ("eV", "meV", "Ha", "cm-1", "Thz")

        Returns:
            A PowderIntensity with the parallel, perpendicular and total components of the powder
            intensities. Each one is an array with length n modes.
        """

        if non_anal_dir is None:
            w = self.phfreqs
            sus = self.susceptibility
        else:
            w = self.non_anal_phfreqs[non_anal_dir]
            sus = self.non_anal_susceptibility[non_anal_dir]

        g0 = np.trace(sus, axis1=1, axis2=2)**2 / 3
        g1 = ((sus[:, 0, 1] - sus[:, 1, 0])**2 + (sus[:, 0, 2] - sus[:, 2, 0])**2 + (sus[:, 2, 1] - sus[:, 1, 2])**2) / 2
        g2 = ((sus[:, 0, 1] + sus[:, 1, 0])**2 + (sus[:, 0, 2] + sus[:, 2, 0])**2 + (sus[:, 2, 1] + sus[:, 1, 2])**2) / 2 + \
             ((sus[:, 0, 0] - sus[:, 1, 1])**2 + (sus[:, 0, 0] - sus[:, 2, 2])**2 + (sus[:, 1, 1] - sus[:, 2, 2])**2) / 3

        laser_freq = laser_freq / abu.phfactor_ev2units(units)

        c = self._get_prefactor(w=w, temp=temp, laser_freq=laser_freq)

        paral = c * (10 * g0 + 4 * g2)
        perp = c * (5 * g1 + 3 * g2)
        tot = paral + perp
        if relative:
            m = tot.max()
            paral /= m
            perp /= m
            tot /= m

        return PowderIntensity(paral, perp, tot)

    def get_powder_lorentz_intensity(self, temp, laser_freq, width, non_anal_dir=None, min_freq=None,
                                     max_freq=None, num=1000, relative=False, units="eV") -> PowderIntensity:
        """
        Calculates the broadened Raman intensities in arbitrary units integrated over all possible
        orientation to reproduce the powder measurements for frequencies in an interval. It is possible to
        use the susceptibilities from the transverse modes only or to specify one of the directions with non
        analytical contributions.

        Args:
            temp: temperature in K.
            laser_freq: frequency of the incident laser. The units are determined the "units"
                argument.
            width: the width of the Lorentz distribution. The units are determined the "units"
                argument.
            non_anal_dir: index of the direction along which the non analytical contribution
                has been calculated. Corresponds to the indices in the non_anal_directions attribute.
            min_freq: minimum frequency considered. If None it will be given by the minimum
                frequency with non zero intensities minus 10 times the width of the distribution.
                If given the units are determined by the "units" argument.
            max_freq: maximum frequency considered. If None it will be given by the maximum
                frequency with non zero intensities plus 10 times the width of the distribution.
                If given the units are determined by the "units" argument.
            num: number of frequencies in the interval (min_freq, max_freq).
            relative: if True the intensities will be rescaled so that the largest value of the
                total intensity is 1.
            units: the units in which the input and the output frequencies will be given.
                Possible values in ("eV", "meV", "Ha", "cm-1", "Thz")

        Returns:
            A PowderIntensity with the parallel, perpendicular and total components of the powder
            intensities. Each one is a Function1D with "num" points.
        """

        pi = self.get_powder_intensity(temp=temp, laser_freq=laser_freq, non_anal_dir=non_anal_dir, units=units)

        freqs, lorentz = self._get_lorentz_freqs_and_factor(intensity=pi.tot, non_anal_dir=non_anal_dir, min_freq=min_freq,
                                                            max_freq=max_freq, num=num, width=width, units=units)

        lpi = np.array([i.dot(lorentz) for i in pi])
        if relative:
            lpi /= lpi[2].max()

        # now convert the frequencies to the desired units for the output
        x = freqs * abu.phfactor_ev2units(units)

        return PowderIntensity(*(Function1D(x, y) for y in lpi))

    @add_fig_kwargs
    def plot_intensity(self, temp, laser_freq, width, value, non_anal_dir=None, min_freq=None, max_freq=None,
                       num=1000, relative=False, units="eV", ax=None, 
                       plot_phfreqs=False, **kwargs) -> Figure:
        """
        Plot one representation of the broadened Raman intensities.

        Args:
            temp: temperature in K.
            laser_freq: frequency of the incident laser. The units are determined according to
                the "units" argument.
            width: the width of the Lorentz distribution. The units are determined the "units"
                argument. If None or 0 a plot of only the frequencies for each mode will be given.
            value: a string describing the value that should be plotted. Can be "powder" or
                a string of the type "xz" with the polarization of the incoming and outgoing
                phonon. All the combinations of "x", "y" and "z" are accepted.
            non_anal_dir: index of the direction along which the non analytical contribution
                has been calculated. Corresponds to the indices in the non_anal_directions attribute.
            min_freq: minimum frequency considered. If None it will be given by the minimum
                frequency with non zero intensities minus 10 times the width of the distribution.
                If given the units are determined by the "units" argument.
            max_freq: maximum frequency considered. If None it will be given by the maximum
                frequency with non zero intensities plus 10 times the width of the distribution.
                If given the units are determined by the "units" argument.
            num: number of frequencies in the interval (min_freq, max_freq).
            relative: if True the intensities will be rescaled so that the largest value of the
                total intensity is 1.
            units: the units in which the input and the output frequencies will be given.
                Possible values in ("eV", "meV", "Ha", "cm-1", "Thz")
            ax: |matplotlib-Axes| or None if a new figure should be created.
            plot_phfreqs: if True vertical dashed lines are added to the figure for all the
                phonon modes.
            **kwargs: arguments passed to the plot function.

        Returns: |matplotlib-Figure|
        """

        ax, fig, plt = get_ax_fig_plt(ax=ax)

        if width:
            if value == "powder":
                f = self.get_powder_lorentz_intensity(temp=temp, laser_freq=laser_freq, width=width,
                                                      non_anal_dir=non_anal_dir, min_freq=min_freq, max_freq=max_freq,
                                                      num=num, relative=relative, units=units).tot
            else:
                pol_in = value[0]
                pol_out = value[1]
                f = self.get_lorentz_intensity(temp=temp, laser_freq=laser_freq, width=width, non_anal_dir=non_anal_dir,
                                               min_freq=min_freq, max_freq=max_freq, num=num, relative=relative,
                                               units=units, pol_in=pol_in, pol_out=pol_out)

            f.plot(ax=ax, **kwargs)

            if plot_phfreqs:
                if non_anal_dir is None:
                    w = self.phfreqs
                else:
                    w = self.non_anal_phfreqs[non_anal_dir]

                w = w * abu.phfactor_ev2units(units)

                min_freq = f.mesh[0]
                max_freq = f.mesh[-1]

                for wi in w:
                    if min_freq < wi < max_freq:
                        ax.axvline(x=wi, ls="--", color="k", lw=0.5)

        else:
            if value == "powder":
                ri = self.get_powder_intensity(temp=temp, laser_freq=laser_freq, non_anal_dir=non_anal_dir,
                                               relative=relative, units=units)

                i = ri.tot
            else:
                if len(value) != 2:
                    raise ValueError("The value should contain the ingoing and outgoing polarizations.")
                pol_in = value[0]
                pol_out = value[1]
                i = self.get_modes_intensities(temp=temp, laser_freq=laser_freq, non_anal_dir=non_anal_dir,
                                               relative=relative, units=units, pol_in=pol_in, pol_out=pol_out)

            if non_anal_dir is None:
                w = self.phfreqs * abu.phfactor_ev2units(units)
            else:
                w = self.non_anal_phfreqs[non_anal_dir] * abu.phfactor_ev2units(units)

            ax.stem(w, i, **kwargs)

        return fig



#!/usr/bin/env python
r"""
Born effective charges and dielectric tensors with DFPT
=======================================================

This example shows how to compute the Born effective charges and
the dielectric tensors (e0, einf) of AlAs with AbiPy flows.
We perform multiple calculations by varying the number of k-points
in order to analyze the convergence of the results wrt nkpt.
"""
import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata

from abipy import flowtk


def make_scf_input(ngkpt, paral_kgb=0):
    """
    This function constructs the input file for the GS calculation for a given IBZ sampling.
    """
    # Crystalline AlAs: computation of the second derivative of the total energy
    structure = abidata.structure_from_ucell("AlAs")
    pseudos = abidata.pseudos("13al.981214.fhi", "33as.pspnc")
    gs_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    gs_inp.set_vars(
        nband=4,
        ecut=2.0,
        ngkpt=ngkpt,
        nshiftk=4,
        shiftk=[0.0, 0.0, 0.5,   # This gives the usual fcc Monkhorst-Pack grid
                0.0, 0.5, 0.0,
                0.5, 0.0, 0.0,
                0.5, 0.5, 0.5],
        #shiftk=[0, 0, 0],
        paral_kgb=paral_kgb,
        tolvrs=1.0e-10,
        diemac=9.0,
        ixc=1,      # This is needed because the pseudos have been generated with different XC
        #iomode=3,
    )

    return gs_inp


def build_flow(options):
    """
    Create a `Flow` for phonon calculations. The flow has two works.

    The first work contains a single GS task that produces the WFK file used in DFPT
    Then we have multiple Works that are generated automatically
    in order to compute the dynamical matrix on a [4, 4, 4] mesh.
    Symmetries are taken into account: only q-points in the IBZ are generated and
    for each q-point only the independent atomic perturbations are computed.
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    flow = flowtk.Flow(workdir=options.workdir)

    for ngkpt in [(2, 2, 2), (4, 4, 4), (8, 8, 8)]:
        # Build input for GS calculation with different k-meshes
        scf_input = make_scf_input(ngkpt=ngkpt)
        flow.register_scf_task(scf_input, append=True)

    for scf_task in flow[0]:
        bec_work = flowtk.BecWork.from_scf_task(scf_task)
        flow.register_work(bec_work)

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_becs_and_epsilon_vs_kpts.py -s
#
# Use:
#
#   abirun.py flow_becs_and_epsilon_vs_kpts/ listext DDB
#
# to list all the DDB files produced by the flow.
#
# Now use abicomp.py to create a DdbRobot to analyze the DDB files produced by the 3 Works:
#
#    abicomp.py ddb flow_becs_and_epsilon_vs_kpts/w*/outdata/out_DDB
#
# the inside the ipython terminal use the anacompare methods to analyze the results. e.g.
#
# .. code-block:: ipython
#
#     In [1]: r = robot.anacompare_epsinf()
#     In [2]: r.df
#     Out[2]:
#               xx         yy         zz   yz   xz   xy  formula  chneut  \
#     0  13.584082  13.584082  13.584082  0.0  0.0  0.0  Al1 As1       1
#     0   9.268425   9.268425   9.268425  0.0  0.0  0.0  Al1 As1       1
#     0   8.873178   8.873178   8.873178  0.0  0.0  0.0  Al1 As1       1
#



#!/usr/bin/env python
r"""
Convergence studies of e-DOS wrt ngkpt
======================================

This examples shows how to build a Flow to compute the
band structure and the electron DOS of MgB2 with different k-point samplings.
"""

import os
import sys
import abipy.data as abidata
import abipy.abilab as abilab
from abipy import flowtk


def make_scf_nscf_inputs(structure, pseudos, paral_kgb=1):
    """return GS, NSCF (band structure), and DOSes input."""

    multi = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=5)

    # Global variables
    multi.set_vars(
        ecut=10,
        nband=11,
        timopt=-1,
        occopt=4,    # Marzari smearing
        tsmear=0.03,
        paral_kgb=paral_kgb,
   )

    # Dataset 1 (GS run)
    multi[0].set_kmesh(ngkpt=[8, 8, 8], shiftk=structure.calc_shiftk())
    multi[0].set_vars(tolvrs=1e-6)

    # Dataset 2 (NSCF Band Structure)
    multi[1].set_kpath(ndivsm=6)
    multi[1].set_vars(tolwfr=1e-12)

    # Dos calculations with increasing k-point sampling.
    for i, nksmall in enumerate([4, 8, 16]):
        multi[i+2].set_vars(
            iscf=-3,   # NSCF calculation
            ngkpt=structure.calc_ngkpt(nksmall),
            shiftk=[0.0, 0.0, 0.0],
            tolwfr=1.0e-10,
        )

    # return GS, NSCF (band structure), DOSes input.
    return multi.split_datasets()


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    structure = abidata.structure_from_ucell("MgB2")

    # Get pseudos from a table.
    table = abilab.PseudoTable(abidata.pseudos("12mg.pspnc", "5b.pspnc"))
    pseudos = table.get_pseudos_for_structure(structure)

    inputs = make_scf_nscf_inputs(structure, pseudos)
    scf_input, nscf_input, dos_inputs = inputs[0], inputs[1], inputs[2:]

    return flowtk.bandstructure_flow(options.workdir, scf_input, nscf_input,
                                     dos_inputs=dos_inputs, manager=options.manager)


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_mgb2_edoses.py -s
#
# then use:
#
#    abirun.py flow_mgb2_edoes ebands
#
# to get info about the electronic properties:
#
# .. code-block:: shell
#
#       KS electronic bands:
#              nsppol  nspinor  nspden  nkpt  nband  nelect  fermie formula  natom  \
#       w0_t0       1        1       1    40     11     8.0   7.615  Mg1 B2      3
#       w0_t1       1        1       1    97     11     8.0   7.615  Mg1 B2      3
#       w0_t2       1        1       1    15     11     8.0   7.701  Mg1 B2      3
#       w0_t3       1        1       1    80     11     8.0   7.629  Mg1 B2      3
#       w0_t4       1        1       1   432     11     8.0   7.626  Mg1 B2      3
#
#              angle0  angle1  angle2      a      b      c  volume abispg_num  \
#       w0_t0    90.0    90.0   120.0  3.086  3.086  3.523  29.056        191
#       w0_t1    90.0    90.0   120.0  3.086  3.086  3.523  29.056        191
#       w0_t2    90.0    90.0   120.0  3.086  3.086  3.523  29.056        191
#       w0_t3    90.0    90.0   120.0  3.086  3.086  3.523  29.056        191
#       w0_t4    90.0    90.0   120.0  3.086  3.086  3.523  29.056        191
#
#                                                         scheme  occopt  tsmear_ev  \
#       w0_t0  cold smearing of N. Marzari with minimization ...       4      0.816
#       w0_t1  cold smearing of N. Marzari with minimization ...       4      0.816
#       w0_t2  cold smearing of N. Marzari with minimization ...       4      0.816
#       w0_t3  cold smearing of N. Marzari with minimization ...       4      0.816
#       w0_t4  cold smearing of N. Marzari with minimization ...       4      0.816
#
#              bandwidth_spin0  fundgap_spin0  dirgap_spin0 task_class  \
#       w0_t0           12.452          0.031         0.609    ScfTask
#       w0_t1           12.441          0.077         0.399   NscfTask
#       w0_t2           12.368          0.415         1.680   NscfTask
#       w0_t3           12.510          0.069         0.390   NscfTask
#       w0_t4           12.506          0.033         0.283   NscfTask
#
#                                                 ncfile  node_id              status
#       w0_t0  flow_mgb2_edoses/w0/t0/outdata/out_GSR.nc   241032  Completed
#       w0_t1  flow_mgb2_edoses/w0/t1/outdata/out_GSR.nc   241033  Completed
#       w0_t2  flow_mgb2_edoses/w0/t2/outdata/out_GSR.nc   241034  Completed
#       w0_t3  flow_mgb2_edoses/w0/t3/outdata/out_GSR.nc   241035  Completed
#       w0_t4  flow_mgb2_edoses/w0/t4/outdata/out_GSR.nc   241036  Completed
#
# Our main goal is to analyze the convergence of the DOS wrt to the k-point sampling.
# As we know that ``w0_t2``, `w0_t3`` and ``w0_t4`` are DOS calculations, we can
# build a GSR robot for these tasks with:
#
#    abirun.py flow_mgb2_edoses/ ebands -nids=241034,241035,241036
#
# then, inside the ipython shell, one can use:
#
# .. code-block:: ipython
#
#    In [1]: %matplotlib
#    In [2]: robot.combiplot_edos()
#
# to plot the electronic DOS obtained with different number of k-points in the IBZ.
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_mgb2_edoses.png?raw=true
#    :alt: Convergence of electronic DOS in MgB2 wrt k-points.
#



#!/usr/bin/env python
r"""
Delta SCF constrained occupation method calculation, to determine luminescent properties.
=========================================================================================

This example shows how to compute the luminescent properties of Eu doped phosphor.
It uses a 36 atoms cell of SrLiAl3N4. Two non-equivalent Sr sites are availabe for Eu, resulting
in two independent LumiWork. The creation of the supercells is done with make_doped_supercell().

Steps, for each structure:
1) Relaxation in the ground state
2) Relaxation in the excited state, starting from the relaxed ground state. Created at run-time
3) Scf computation in the relaxed/unrelaxed ground/excited state (4 computations).

Even if we use minimal settings, the workflow takes around one hour to run on one core.
Filepaths of the 6 runs are stored in outdata/lumi.json of each work
A quick post-processing is automatically done at the end of a LumiWork
and stored in outdata/Delta_SCF.json of each work, with relevant luminescent properties
(ZPL energy, Stoke Shift, \Delta Q,...), see abipy/lumi/delta_scf.py
"""

import sys
import os
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata
from abipy.core.structure import Structure
from abipy.flowtk.lumi_works import LumiWork

def get_non_eq_sites(structure,replaced_atom):
    ### return a list of positions of non-equivalent sites for the replaced atom. ###
    irred=structure.spget_equivalent_atoms().eqmap # mapping from inequivalent sites to atoms sites
    positions=structure.get_symbol2indices()[replaced_atom] # get indices of the replaced atom

    index_different_sites=[]

    for i in positions:
        if len(irred[i]) != 0:
            index_different_sites.append(irred[i][0])

    return(index_different_sites)


def make_doped_supercell(prim_structure,supercell_size,replaced_atom,dopant_atom):
    #return a list of doped supercell structure, one for each non-equivalent site of the replaced atom
    my_structure=prim_structure.copy()
    my_structure.make_supercell(supercell_size)

    list_ineq_pos=get_non_eq_sites(my_structure,replaced_atom)

    doped_structure_list=[]

    for pos in list_ineq_pos:
        final_structure=my_structure.copy()
        final_structure.replace(pos,dopant_atom)
        doped_structure_list.append(final_structure)

    return doped_structure_list


def scf_inp(structure):
    pseudos = abidata.pseudos("Eu.xml", "Sr.xml","Al.xml","N.xml","Li.xml")

    gs_scf_inp = abilab.AbinitInput(structure=structure, pseudos=pseudos)
    gs_scf_inp.set_vars(ecut=10,
                        pawecutdg=20,
                        chksymbreak=0,
                        diemac=5,
                        prtwf=0,
                        nstep=300,
                        toldfe=1e-10,
                        chkprim=0,
                        nbdbuf=5 # help convergence
                    )


    # Set DFT+U and spinat parameters according to chemical symbols.
    #symb2spinat = {"Eu": [0, 0, 7]}
    #symb2luj = {"Eu": {"lpawu": 3, "upawu": 7, "jpawu": 0.7}}

    #gs_scf_inp.set_usepawu(usepawu=1, symb2luj=symb2luj)
    #gs_scf_inp.set_spinat_from_symbols(symb2spinat, default=(0, 0, 0))

    ### Setting of the occupations, for spin up-dn in the ground/excited state
    ### Only valid for Eu doped
    n_val = gs_scf_inp.num_valence_electrons
    n_cond = round(15)

    spin_up_gs = f"\n{int((n_val - 7) / 2)}*1 7*1 {n_cond}*0"
    spin_up_ex = f"\n{int((n_val - 7) / 2)}*1 6*1 0 1 {n_cond - 1}*0"
    spin_dn = f"\n{int((n_val - 7) / 2)}*1 7*0 {n_cond}*0"

    nsppol = 2
    shiftk = [0, 0, 0]
    ngkpt = [1, 1, 1]

    # Build SCF input for the ground state configuration.
    gs_scf_inp.set_kmesh_nband_and_occ(ngkpt, shiftk, nsppol, [spin_up_gs, spin_dn])
    # Build SCF input for the excited configuration.
    exc_scf_inp = gs_scf_inp.deepcopy()
    exc_scf_inp.set_kmesh_nband_and_occ(ngkpt, shiftk, nsppol, [spin_up_ex, spin_dn])

    return gs_scf_inp,exc_scf_inp



def relax_kwargs():

    # Dictionary with input variables to be added for performing structural relaxations.
    relax_kwargs = dict(
        ecutsm=0.5,
        toldff=1e-4, # TOO HIGH, just for testing purposes.
        tolmxf=1e-3, # TOO HIGH, just for testing purposes.
        ionmov=2,
        chkdilatmx=0,
    )

    relax_kwargs_gs=relax_kwargs.copy()
    relax_kwargs_gs['optcell']=0 # in the ground state, allow relaxation of the cell

    relax_kwargs_ex=relax_kwargs.copy()
    relax_kwargs_ex['optcell']=0 # in the excited state, no relaxation of the cell

    return relax_kwargs_gs, relax_kwargs_ex


def build_flow(options):

    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    flow = flowtk.Flow(options.workdir, manager=options.manager)

    #Construct the two structures (2 non-eq. sites for Sr) from the primitive cell of SLA (SrAlLi3N4)

    #prim_structure=structure.Structure.from_file('SLA_prim.cif')
    prim_structure=Structure.from_file(abidata.cif_file("SLA_prim.cif"))
    supercell_matrix=[1,1,1]  # Too small, just for test
    strus=prim_structure.make_doped_supercells(supercell_matrix,'Sr','Eu')


    ####### Delta SCF part of the flow #######

    # Create one LumiWork per structure
    for stru in strus:
        gs_scf_inp, exc_scf_inp = scf_inp(stru)
        relax_kwargs_gs, relax_kwargs_ex = relax_kwargs()
        lumi_work=LumiWork.from_scf_inputs(gs_scf_inp, exc_scf_inp, relax_kwargs_gs, relax_kwargs_ex)
        flow.register_work(lumi_work)

    return flow



# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == '__main__':
    sys.exit(main())




# coding: utf-8
"""Tools for the submission of Tasks."""
from __future__ import annotations

import abc
import os
import time
import datetime
import pandas as pd
import apscheduler

from collections import deque
from io import StringIO
from queue import Queue, Empty
from typing import Optional
from shutil import which
from monty.io import get_open_fds
from monty.string import boxed, is_string
from monty.collections import AttrDict #, dict2namedtuple
from monty.termcolor import cprint
from monty.functools import lazy_property
from abipy.tools.iotools import yaml_safe_load, ask_yesno
from abipy.tools.typing import TYPE_CHECKING
from .utils import as_bool

import logging
logger = logging.getLogger(__name__)

#try:
#    has_sched_v3 = apscheduler.version >= "3.0.0"
#except AttributeError:
#    has_sched_v3 = False

if TYPE_CHECKING:  # needed to avoid circular imports
    from .tasks import Task
    #from .works import Work
    from .flows import Flow


__all__ = [
    "ScriptEditor",
    "PyLauncher",
    "PyFlowScheduler",
    "MultiFlowScheduler",
]


def straceback() -> str:
    """Returns a string with the traceback."""
    import traceback
    return traceback.format_exc()



class ScriptEditor:
    """
    Simple editor to simplify the writing of shell scripts
    """

    _shell = '/bin/bash'

    def __init__(self):
        self._lines = []

    @property
    def shell(self) -> str:
        return self._shell

    def _add(self, text, pre="") -> None:
        if is_string(text):
            self._lines.append(pre + text)
        else:
            self._lines.extend([pre + t for t in text])

    def reset(self) -> None:
        """Reset the editor."""
        try:
            del self._lines
        except AttributeError:
            pass

    def shebang(self) -> None:
        """Adds the shebang line."""
        self._lines.append('#!' + self.shell)

    def declare_var(self, key: str, val: str) -> None:
        """Declare a env variable. If val is None the variable is unset."""
        if val is not None:
            line = "export " + key + '=' + str(val)
        else:
            line = "unset " + key

        self._add(line)

    def declare_vars(self, d: dict) -> None:
        """Declare the variables defined in the dictionary d."""
        for k, v in d.items():
            self.declare_var(k, v)

    def export_envar(self, key: str, val: str) -> None:
        """Export an environment variable."""
        line = "export " + key + "=" + str(val)
        self._add(line)

    def export_envars(self, env: dict) -> None:
        """Export the environment variables contained in the dict env."""
        for k, v in env.items():
            self.export_envar(k, v)

    def add_emptyline(self) -> None:
        """Add an empty line."""
        self._add("", pre="")

    def add_comment(self, comment: str) -> None:
        """Add a comment"""
        self._add(comment, pre="# ")

    def load_modules(self, modules: list[str]) -> None:
        """Load the list of specified modules."""
        for module in modules:
            self.load_module(module)

    def load_module(self, module: str) -> None:
        self._add('module load ' + module + " 2>> mods.err")

    def add_line(self, line: str) -> None:
        self._add(line)

    def add_lines(self, lines: list[str]) -> None:
        self._add(lines)

    def get_script_str(self, reset=True) -> str:
        """Returns a string with the script and reset the editor if reset is True"""
        s = "\n".join(l for l in self._lines)
        if reset:
            self.reset()
        return s


class PyLauncherError(Exception):
    """Error class for PyLauncher."""


class PyLauncher:
    """
    This object handle the submission of the tasks contained in a |Flow|.
    """

    Error = PyLauncherError

    def __init__(self, flow: Flow, **kwargs) -> None:
        """
        Initialize the object

        Args:
            flow: |Flow| object
            max_njobs_inqueue: The launcher will stop submitting jobs when the
                number of jobs in the queue is >= Max number of jobs
        """
        self.flow = flow
        self.max_njobs_inqueue = kwargs.get("max_njobs_inqueue", 200)

        #self.flow.check_pid_file()

    def single_shot(self):
        """
        Run the first :class:`Task` than is ready for execution.

        Returns:
            Number of jobs launched.
        """
        num_launched = 0

        # Get the tasks that can be executed in each workflow.
        tasks = []
        for work in self.flow:
            try:
                task = work.fetch_task_to_run()

                if task is not None:
                    tasks.append(task)
                else:
                    # No task found, this usually happens when we have dependencies.
                    # Beware of possible deadlocks here!
                    logger.debug("No task to run! Possible deadlock")

            except StopIteration:
                logger.info("All tasks completed.")

        # Submit the tasks and update the database.
        if tasks:
            tasks[0].start()
            num_launched += 1

            self.flow.pickle_dump()

        return num_launched

    def rapidfire(self, max_nlaunch=-1, max_loops=1, sleep_time=5):
        """
        Keeps submitting `Tasks` until we are out of jobs or no job is ready to run.

        Args:
            max_nlaunch: Maximum number of launches. default: no limit.
            max_loops: Maximum number of loops
            sleep_time: seconds to sleep between rapidfire loop iterations

        Returns:
            The number of tasks launched.
        """
        num_launched, do_exit, launched = 0, False, []

        for count in range(max_loops):
            if do_exit: break
            if count > 0: time.sleep(sleep_time)

            tasks = self.fetch_tasks_to_run()

            # I don't know why but we receive duplicated tasks.
            if any(task in launched for task in tasks):
                logger.critical("numtasks %d already in launched list:\n%s" % (len(tasks), launched))

            # Preventive test.
            tasks = [t for t in tasks if t not in launched]

            if not tasks: continue

            for task in tasks:
                fired = task.start()
                if fired:
                    launched.append(task)
                    num_launched += 1

                if num_launched >= max_nlaunch > 0:
                    logger.info('num_launched >= max_nlaunch, breaking submission loop')
                    do_exit = True
                    break

        # Update the database.
        self.flow.pickle_dump()

        return num_launched

    def fetch_tasks_to_run(self) -> list[Task]:
        """
        Return the list of tasks that can be submitted.
        Empty list if no task has been found.
        """
        tasks_to_run = []

        for work in self.flow:
            tasks_to_run.extend(work.fetch_alltasks_to_run())

        return tasks_to_run


class PyFlowSchedulerError(Exception):
    """Exceptions raised by `PyFlowScheduler`."""


class BaseScheduler(metaclass=abc.ABCMeta):
    """
    This object schedules the submission of the tasks in a |Flow|.
    There are two types of errors that might occur during the execution of the jobs:

        #. Python exceptions
        #. Errors in the ab-initio code

    Python exceptions are easy to detect and are usually due to a bug in the python code or random errors such as IOError.
    The set of errors in the ab-initio is much much broader. It includes wrong input data, segmentation
    faults, problems with the resource manager, etc. The flow tries to handle the most common cases
    but there's still a lot of room for improvement.
    Note, in particular, that `PyFlowScheduler` will shutdown automatically in the following cases:

        #. The number of python exceptions is > max_num_pyexcs

        #. The number of task errors (i.e. the number of tasks whose status is S_ERROR) is > max_num_abierrs

        #. The number of jobs launched becomes greater than (`safety_ratio` * total_number_of_tasks).

        #. The scheduler will send an email to the user (specified by `mailto`) every `remindme_s` seconds.
           If the mail cannot be sent, the scheduler will automatically shutdown.
           This check prevents the scheduler from being trapped in an infinite loop.
    """
    # Configuration file.
    YAML_FILE = "scheduler.yml"

    USER_CONFIG_DIR = os.path.join(os.path.expanduser("~"), ".abinit", "abipy")

    Error = PyFlowSchedulerError

    @classmethod
    def autodoc(cls) -> str:
        """Return string with scheduler options."""
        i = cls.__init__.__doc__.index("Args:")
        return cls.__init__.__doc__[i+5:]

    def __init__(self, **kwargs):
        """
        Args:
            weeks: number of weeks to wait (DEFAULT: 0).
            days: number of days to wait (DEFAULT: 0).
            hours: number of hours to wait (DEFAULT: 0).
            minutes: number of minutes to wait (DEFAULT: 0).
            seconds: number of seconds to wait (DEFAULT: 0).
            mailto: The scheduler will send an email to `mailto` every `remindme_s` seconds.
                (DEFAULT: None i.e. not used).
            verbose: (int) verbosity level. (DEFAULT: 0)
            use_dynamic_manager: "yes" if the |TaskManager| must be re-initialized from
                file before launching the jobs. (DEFAULT: "no")
            max_njobs_inqueue: Limit on the number of jobs that can be present in the queue. (DEFAULT: 200)
            max_ncores_used: Maximum number of cores that can be used by the scheduler.
            remindme_s: The scheduler will send an email to the user specified
                by `mailto` every `remindme_s` seconds. (int, DEFAULT: 1 day).
            max_num_pyexcs: The scheduler will exit if the number of python exceptions is > max_num_pyexcs
                (int, DEFAULT: 0)
            max_num_abierrs: The scheduler will exit if the number of errored tasks is > max_num_abierrs
                (int, DEFAULT: 0)
            safety_ratio: The scheduler will exits if the number of jobs launched becomes greater than
               `safety_ratio` * total_number_of_tasks_in_flow. (int, DEFAULT: 5)
            max_nlaunches: Maximum number of tasks launched in a single iteration of the scheduler.
                (DEFAULT: -1 i.e. no limit)
            debug: Debug level. Use 0 for production (int, DEFAULT: 0)
            fix_qcritical: "yes" if the launcher should try to fix QCritical Errors (DEFAULT: "no")
            rmflow: If "yes", the scheduler will remove the flow directory if the calculation
                completed successfully. (DEFAULT: "no")
            killjobs_if_errors: "yes" if the scheduler should try to kill all the running jobs
                before exiting due to an error. (DEFAULT: "yes")
        """
        self.init_kwargs = kwargs.copy()

        # Options passed to the apscheduler scheduler.
        self.sched_options = AttrDict(
            weeks=kwargs.pop("weeks", 0),
            days=kwargs.pop("days", 0),
            hours=kwargs.pop("hours", 0),
            minutes=kwargs.pop("minutes", 0),
            seconds=kwargs.pop("seconds", 0),
        )
        if all(not v for v in self.sched_options.values()):
            raise self.Error("Please specify at least one option among: seconds, minutes, hours ...")

        self.mailto = kwargs.pop("mailto", None)
        self.verbose = int(kwargs.pop("verbose", 0))
        self.use_dynamic_manager = as_bool(kwargs.pop("use_dynamic_manager", False))
        self.max_njobs_inqueue = kwargs.pop("max_njobs_inqueue", 200)
        self.max_ncores_used = kwargs.pop("max_ncores_used", None)

        self.remindme_s = float(kwargs.pop("remindme_s", 1 * 24 * 3600))
        self.max_num_pyexcs = int(kwargs.pop("max_num_pyexcs", 0))
        self.max_num_abierrs = int(kwargs.pop("max_num_abierrs", 0))
        self.safety_ratio = int(kwargs.pop("safety_ratio", 5))
        #self.max_etime_s = kwargs.pop("max_etime_s", 14 * 3600)
        self.max_nlaunches = kwargs.pop("max_nlaunches", -1)
        self.debug = kwargs.pop("debug", 0)
        self.fix_qcritical = as_bool(kwargs.pop("fix_qcritical", False))
        self.rmflow = as_bool(kwargs.pop("rmflow", False))
        self.killjobs_if_errors = as_bool(kwargs.pop("killjobs_if_errors", True))

        # TODO: Add abinit_options, anaddb_options, exec_options

        if kwargs:
            raise self.Error("Unknown arguments `%s`" % str(kwargs))

        # Register the callaback in the scheduler
        #if has_sched_v3:
        logger.warning("Using scheduler v >= 3.0.0")
        from apscheduler.schedulers.blocking import BlockingScheduler
        self.sched = BlockingScheduler()
        self.sched.add_job(self.callback, "interval", **self.sched_options)
        #else:
        #    from apscheduler.scheduler import Scheduler
        #    self.sched = Scheduler(standalone=True)
        #    self.sched.add_interval_job(self.callback, **self.sched_options)

        self.nlaunch = 0
        self.num_reminders = 1
        self.start_time = None

        # Used to keep track of the exceptions raised while the scheduler is running
        self.exceptions = deque(maxlen=self.max_num_pyexcs + 10)

        # Used to push additional info during the execution.
        self.history = deque(maxlen=200)

    @classmethod
    def from_file(cls, filepath: str) -> BaseScheduler:
        """Read the configuration parameters from a Yaml file."""
        with open(filepath, "rt") as fh:
            return cls(**yaml_safe_load(fh))

    @classmethod
    def from_string(cls, s: str) -> BaseScheduler:
        """Create an istance from string s containing a YAML dictionary."""
        stream = StringIO(s)
        stream.seek(0)
        return cls(**yaml_safe_load(stream))

    @classmethod
    def from_user_config(cls) -> BaseScheduler:
        """
        Initialize the :class:`PyFlowScheduler` from the YAML file 'scheduler.yml'.
        Search first in the working directory and then in the configuration directory of abipy.

        Raises:
            `RuntimeError` if file is not found.
        """
        # Try in the current directory.
        path = os.path.join(os.getcwd(), cls.YAML_FILE)
        if os.path.exists(path):
            return cls.from_file(path)

        # Try in the configuration directory.
        path = os.path.join(cls.USER_CONFIG_DIR, cls.YAML_FILE)
        if os.path.exists(path):
            return cls.from_file(path)

        raise cls.Error(f"""
Cannot locate {cls.YAML_FILE} neither in current directory nor in {path}.

An example of scheduler.yml is given below:

{cls.get_default_scheduler()}
""")

    @classmethod
    def get_default_scheduler(cls) -> str:
        return """
weeks: 0 # number of weeks to wait (DEFAULT: 0).
days: 0 # number of days to wait (DEFAULT: 0).
hours: 0 # number of hours to wait (DEFAULT: 0).
minutes: 0 # number of minutes to wait (DEFAULT: 0).
seconds: 0 # number of seconds to wait (DEFAULT: 0).
mailto: null # The scheduler will send an email to `mailto` every `remindme_s` seconds. (DEFAULT: None i.e. not used).
verbose: 0 # (int) verbosity level. (DEFAULT: 0)
use_dynamic_manager: no # "yes" if the |TaskManager| must be re-initialized from file before launching the jobs. (DEFAULT: "no")
max_njobs_inqueue: 200 # Limit on the number of jobs that can be present in the queue. (DEFAULT: 200)
max_ncores_used: null # Maximum number of cores that can be used by the scheduler. (DEFAULT: None)
remindme_s: 1 # The scheduler will send an email to the user specified by `mailto` every `remindme_s` seconds. (int, DEFAULT: 1 day).
max_num_pyexcs: 0 # The scheduler will exit if the number of python exceptions is > max_num_pyexcs (int, DEFAULT: 0)
max_num_abierrs: 0 # The scheduler will exit if the number of errored tasks is > max_num_abierrs (int, DEFAULT: 0)
safety_ratio: 5 # The scheduler will exits if the number of jobs launched becomes greater than safety_ratio` * total_number_of_tasks_in_flow. (int, DEFAULT: 5)
max_nlaunches: -1 # Maximum number of tasks launched in a single iteration of the scheduler. (DEFAULT: -1 i.e. no limit)
debug: 0 # Debug level. Use 0 for production (int, DEFAULT: 0)
fix_qcritical: no # "yes" if the launcher should try to fix QCritical Errors (DEFAULT: "no")
rmflow: no # If "yes", the scheduler will remove the flow directory if the calculation completed successfully. (DEFAULT: "no")
killjobs_if_errors: yes # "yes" if the scheduler should try to kill all the running jobs before exiting due to an error. (DEFAULT: "yes")
"""

    def __str__(self) -> str:
        """String representation."""
        lines = [self.__class__.__name__ + ", Pid: %d" % self.pid]
        app = lines.append
        app("Scheduler options:\n%s" % str(self.sched_options))

        return "\n".join(lines)

    @abc.abstractmethod
    def add_flow(self, flow: Flow, **kwargs) -> None:
        """
        Add a flow to the scheduler.
        """

    @abc.abstractmethod
    def start(self) -> int:
        """
        Starts the scheduler. Returns 0 if success.
        """

    @abc.abstractmethod
    def callback(self):
        """The function that will be executed by the scheduler."""

    @lazy_property
    def pid(self) -> int:
        """The pid of the process associated to the scheduler."""
        return os.getpid()

    @property
    def num_excs(self) -> int:
        """Number of exceptions raised so far."""
        return len(self.exceptions)

    def get_delta_etime(self):
        """Returns a `timedelta` object representing with the elapsed time."""
        return datetime.timedelta(seconds=(time.time() - self.start_time))

    def cancel_jobs_if_requested(self, flow: Flow) -> None:
        """Cancel all the jobs in the flow"""
        if not self.killjobs_if_errors: return
        cprint("killjobs_if_errors set to 'yes'. Killing jobs before aborting the flow.", "yellow")

        try:
            num_cancelled = 0
            for task in flow.iflat_tasks():
                num_cancelled += task.cancel()
            cprint("Killed %d tasks" % num_cancelled, "yellow")
        except Exception as exc:
            cprint("Exception while trying to kill jobs:\n%s" % str(exc), "red")

    def _accept_flow(self, flow: Flow) -> None:

        # Check if we are already using a scheduler to run this flow
        flow.check_pid_file()
        flow.set_spectator_mode(False)

        # Build dirs and files (if not yet done)
        flow.build()

        errors = flow.look_before_you_leap()
        if errors:
            raise self.Error(str(errors))

    def restart_unconverged(self, flow: Flow, max_nlaunch, excs):
        if max_nlaunch <= 0: return 0

        for task in flow.unconverged_tasks:
            try:
                logger.info("Trying to restart task: `%s`" % repr(task))
                fired = task.restart()
                if fired:
                    self.nlaunch += 1
                    max_nlaunch -= 1
                    if max_nlaunch <= 0:
                        logger.info("Restart: too many jobs in the queue, returning")
                        flow.pickle_dump()
                        return 0

            except task.RestartError:
                excs.append(straceback())

        return max_nlaunch

    def try_to_fix_flow(self, flow: Flow) -> None:

        # Temporarily disabled by MG because I don't know if fix_critical works after the
        # introduction of the new qadapters
        # reenabled by MsS disable things that do not work at low level
        # fix only prepares for restarting, and sets to ready
        if self.fix_qcritical:
            nfixed = flow.fix_queue_critical()
            if nfixed: print("Fixed %d QCritical error(s)" % nfixed)

        nfixed = flow.fix_abicritical()
        if nfixed: print("Fixed %d AbiCritical error(s)" % nfixed)

    def check_deadlocks(self, flow: Flow) -> list[str]:
        err_lines = []

        g = flow.find_deadlocks()
        if g.deadlocked:
            # Check the flow again so that status are updated.
            flow.check_status()

            g = flow.find_deadlocks()
            print("deadlocked:", len(g.deadlocked), ", runnables:", len(g.runnables), ", running:", len(g.running))
            if g.deadlocked and not g.runnables and not g.running:
                err_lines.append("No runnable job with deadlocked tasks:\n%s." % str(g.deadlocked))

        if not g.runnables and not g.running:
            # Check the flow again so that status are updated.
            flow.check_status()
            g = flow.find_deadlocks()
            if not g.runnables and not g.running:
                err_lines.append("No task is running and cannot find other tasks to submit.")

        return err_lines

    def send_email(self, msg: str, tag=None) -> int:
        """
        Send an e-mail before completing the shutdown.
        Returns 0 if success. Relies on _send_email method provided by subclass.
        """
        try:
            return self._send_email(msg, tag)
        except Exception:
            self.exceptions.append(straceback())
            return -2


class PyFlowScheduler(BaseScheduler):

    @property
    def pid_file(self) -> str:
        """
        Absolute path of the file with the pid.
        The file is located in the workdir of the flow
        """
        return self._pid_file

    @property
    def flow(self) -> Flow:
        """`Flow`."""
        try:
            return self._flow
        except AttributeError:
            return None

    def add_flow(self, flow: Flow) -> None:
        """
        Add a flow to the scheduler.
        """
        if hasattr(self, "_flow"):
            raise self.Error(f"Only one Flow can be added to a {self.__class__.__name__} scheduler.")

        self._accept_flow(flow)

        with open(flow.pid_file, "wt") as fh:
            fh.write(str(self.pid))

        self._pid_file = flow.pid_file
        self._flow = flow

    def start(self) -> int:
        """
        Starts the scheduler. Returns 0 if success.
        This method blocks until the flow is completed or some exception is raised.
        """
        self.history.append("Started on %s" % time.asctime())
        self.start_time = time.time()

        flow = self.flow
        if flow is not None:
            # Try to run the job immediately.
            # If something goes wrong return without initializing the scheduler.
            self._runem_all()

            if self.exceptions:
                self.cleanup()
                self.send_email(msg="Error while trying to run the flow for the first time!\n %s" % self.exceptions)
                return 1

        # Start the scheduler loop.
        try:
            self.sched.start()
            return 0

        except KeyboardInterrupt:
            self.shutdown(msg="KeyboardInterrupt from user")
            if ask_yesno("Do you want to cancel all the jobs in the queue? [Y/n]"):
                print("Number of jobs cancelled:", flow.cancel())

            flow.pickle_dump()
            return -1

    def _runem_all(self) -> None:
        """
        This function checks the status of all tasks,
        tries to fix tasks that went unconverged, abicritical, or queuecritical
        and tries to run all the tasks that can be submitted.
        """
        excs = []
        flow = self.flow
        if flow is None: return

        if self.use_dynamic_manager:
            # Allow to change the manager at run-time
            from .tasks import TaskManager
            flow.change_manager(TaskManager.from_user_config())

        # Here we just count the number of tasks in the flow that are in RUNNING or SUBMITTED status.
        # This logic clearly breaks down if there are multiple schedulers running on the same machine
        # but it's easy to implement without having to contact the resource manager
        # by calling `flow.get_njobs_in_queue()` that is quite expensive (and the sysadmin won't be happy!)
        nqjobs = (len(list(flow.iflat_tasks(status=flow.S_RUN))) +
                  len(list(flow.iflat_tasks(status=flow.S_SUB))))

        if nqjobs >= self.max_njobs_inqueue:
            print(f"Too many jobs in the queue: {nqjobs} >= {self.max_njobs_inqueue}.\n",
                  "No job will be submitted.")
            flow.check_status(show=False)
            return

        max_nlaunch = self.max_njobs_inqueue - nqjobs if self.max_nlaunches == -1 else \
                      min(self.max_njobs_inqueue - nqjobs, self.max_nlaunches)

        # check status.
        flow.check_status(show=False)

        # This check is not perfect, we should make a list of tasks to submit
        # and then select a subset so that we don't exceeed max_ncores_used
        # Many sections of this code should be rewritten though.
        #if self.max_ncores_used is not None and flow.ncores_used > self.max_ncores_used:
        if self.max_ncores_used is not None and flow.ncores_allocated > self.max_ncores_used:
            print("Cannot exceed max_ncores_used %s" % self.max_ncores_used,
                  ", ncores_allocated:", flow.ncores_allocated)
            return

        # Try to restart unconverged tasks.
        max_nlaunch = self.restart_unconverged(flow, max_nlaunch, excs)
        if max_nlaunch <= 0: return

        self.try_to_fix_flow(flow)

        # Update the pickle file.
        flow.pickle_dump()

        # Submit the tasks that are ready.
        try:
            nlaunch = PyLauncher(flow).rapidfire(max_nlaunch=max_nlaunch, sleep_time=10)
            self.nlaunch += nlaunch
            if nlaunch:
                cprint("[%s] Number of launches: %d" % (time.asctime(), nlaunch), "yellow")

        except Exception:
            excs.append(straceback())

        flow.show_status()

        if excs:
            logger.critical("*** Scheduler exceptions:\n *** %s" % "\n".join(excs))
            self.exceptions.extend(excs)

    def callback(self):
        """The function that will be executed by the scheduler."""
        try:
            return self._callback()
        except Exception:
            # All exceptions raised here will trigger the shutdown!
            s = straceback()
            self.exceptions.append(s)
            # This is useful when debugging
            #self.cancel_jobs_if_requested(self.flow)
            self.shutdown(msg="Exception raised in callback!\n" + s)

    def _callback(self):
        """The actual callback."""
        if self.debug: print(">>>>> _callback: Number of open file descriptors: %s" % get_open_fds())

        self._runem_all()

        flow = self.flow
        all_ok = flow.all_ok

        if all_ok:
            # Mission accomplished. Shutdown the scheduler.
            print("Calling flow.finalize() ...")
            flow.finalize()
            if self.rmflow:
                print("Flow directory will be removed...")
                try:
                    flow.rmtree()
                except Exception:
                    logger.warning("Ignoring exception while trying to remove flow dir.")

            return self.shutdown(msg="All tasks have reached S_OK. Will shutdown the scheduler and exit")

        # Handle failures.
        err_lines = []

        # Shall we send a reminder to the user?
        delta_etime = self.get_delta_etime()

        if delta_etime.total_seconds() > self.num_reminders * self.remindme_s:
            self.num_reminders += 1
            msg = ("Just to remind you that the scheduler with pid %s, flow %s\n has been running for %s " %
                  (self.pid, flow, delta_etime))
            retcode = self.send_email(msg, tag="[REMINDER]")

            if retcode:
                msg += ("\nThe scheduler tried to send an e-mail to remind the user\n" +
                        " but send_email returned %d. Error is not critical though!" % retcode)
                print(msg)

        #if delta_etime.total_seconds() > self.max_etime_s:
        #    err_lines.append("\nExceeded max_etime_s %s. Will shutdown the scheduler and exit" % self.max_etime_s)

        # Too many exceptions. Shutdown the scheduler.
        if self.num_excs > self.max_num_pyexcs:
            msg = "Number of exceptions %s > %s. Will shutdown the scheduler and exit" % (
                self.num_excs, self.max_num_pyexcs)
            err_lines.append(boxed(msg))

        # Paranoid check: disable the scheduler if we have submitted
        # too many jobs (it might be due to some bug or other external reasons
        # such as race conditions between difference callbacks!)
        if self.nlaunch > self.safety_ratio * flow.num_tasks:
            msg = "Too many jobs launched %d. Total number of tasks = %s, Will shutdown the scheduler and exit" % (
                self.nlaunch, flow.num_tasks)
            err_lines.append(boxed(msg))

        # Count the number of tasks with status == S_ERROR.
        if flow.num_errored_tasks > self.max_num_abierrs:
            msg = "Number of tasks with ERROR status %s > %s. Will shutdown the scheduler and exit" % (
                flow.num_errored_tasks, self.max_num_abierrs)
            err_lines.append(boxed(msg))

        # Test for deadlocks.
        errors = self.check_deadlocks(flow)
        if errors:
            err_lines.extend(errors)

        # Something wrong. Quit
        if err_lines:
            self.cancel_jobs_if_requested(flow)
            self.shutdown("\n".join(err_lines))

        return len(self.exceptions)

    def cleanup(self) -> None:
        """Cleanup routine: remove the pid file and save the pickle database"""
        try:
            os.remove(self.pid_file)
        except OSError as exc:
            logger.critical("Could not remove pid_file: %s", exc)

        # Save the final status of the flow.
        self.flow.pickle_dump()

    def shutdown(self, msg: str) -> None:
        """Shutdown the scheduler."""
        flow = self.flow
        all_ok = flow.all_ok

        try:
            self.cleanup()
            self.history.append("Completed on: %s" % time.asctime())
            self.history.append("Elapsed time: %s" % self.get_delta_etime())

            if self.debug:
                print(">>>>> shutdown: Number of open file descriptors: %s" % get_open_fds())

            retcode = self.send_email(msg)
            if self.debug: print("send_mail retcode", retcode)

            # Write file with the list of exceptions:
            if self.exceptions:
                dump_file = os.path.join(flow.workdir, "_exceptions")
                with open(dump_file, "wt") as fh:
                    fh.writelines(self.exceptions)
                    fh.write("Shutdown message:\n%s" % msg)

            lines = []
            app = lines.append
            app("Submitted on: %s" % time.ctime(self.start_time))
            app("Completed on: %s" % time.asctime())
            app("Elapsed time: %s" % str(self.get_delta_etime()))

            if all_ok:
                app("Flow completed successfully")
            else:
                app("Flow %s didn't complete successfully" % repr(flow.workdir))
                app("use `abirun.py FLOWDIR debug` to analyze the problem.")
                app("Shutdown message:\n%s" % msg)

            print("")
            print("\n".join(lines))
            print("")

        finally:
            # Shutdown the scheduler thus allowing the process to exit.
            logger.debug('This should be the shutdown of the scheduler')

            # Unschedule all the jobs before calling shutdown
            #self.sched.print_jobs()
            #if not has_sched_v3:
            #    #self.sched.print_jobs()
            #    for job in self.sched.get_jobs():
            #        self.sched.unschedule_job(job)
            #    self.sched.shutdown()
            #else:
            self.sched.shutdown(wait=False)

            # Uncomment the line below if shutdown does not work!
            #os.system("kill -9 %d" % os.getpid())

    def _send_email(self, msg: str, tag) -> int:
        if self.mailto is None: return -1

        header = msg.splitlines()
        app = header.append
        app("Submitted on: %s" % time.ctime(self.start_time))
        app("Completed on: %s" % time.asctime())
        app("Elapsed time: %s" % str(self.get_delta_etime()))

        # Add the status of the flow.
        flow = self.flow
        app("Number of errored tasks: %d" % flow.num_errored_tasks)
        app("Number of unconverged tasks: %d" % flow.num_unconverged_tasks)
        strio = StringIO()
        strio.writelines("\n".join(header) + 4 * "\n")
        flow.show_status(stream=strio)

        if self.exceptions:
            # Report the list of exceptions.
            strio.writelines(self.exceptions)

        if tag is None:
            tag = " [ALL OK]" if flow.all_ok else " [WARNING]"

        return sendmail(subject=flow.name + tag, text=strio.getvalue(), mailto=self.mailto)


class MultiFlowScheduler(BaseScheduler):

    # TODO: history, logging, shutdown better treatment of exceptions....

    def __init__(self, sqldb_path, **kwargs):
        super().__init__(**kwargs)
        self.flows = []

        self.completed_flows = []
        self.errored_flows = []
        self._errored_flow_ids = []

        self.incoming_flow_queue = Queue()

        import threading
        self._lock = threading.Lock()

        self.sqldb_path = sqldb_path
        self.create_sqldb()

    def add_flow(self, flow: Flow, user_message, priority=None) -> None:
        """
        Add a flow to the scheduler.
        """
        # TODO: Should check the pid file
        self._accept_flow(flow)
        flow.set_user_message(user_message)
        self.incoming_flow_queue.put(flow)

    def register_flow_exception(self, flow_idx, exc) -> None:
        flow = self.flows[flow_idx]
        self.history.append(f"Exception for {repr(flow)}")
        self.history.append(straceback())
        self._errored_flow_ids.append(flow_idx)

    def handle_flow_exception(self) -> None:
        if not self._errored_flow_ids: return
        for idx in self._errored_flow_ids:
            flow = self.flows[idx]
            self.errored_flows.append(flow)

        self.flows = [flow for (idx, flow) in self.flows if idx not in set(self._errored_flow_ids)]
        self._errored_flow_ids = []

    def start(self) -> int:
        """
        Starts the scheduler. Returns 0 if success.
        This method blocks until the flow is completed or some exception is raised.
        """
        self.history.append("Started on %s" % time.asctime())
        self.start_time = time.time()
        self.sched.start()

    # TODO
    #def stop(self):
    #def restart(self):

    def sql_connect(self):
        import sqlite3
        con = sqlite3.connect(self.sqldb_path, check_same_thread=True,
                              detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)
        con.row_factory = sqlite3.Row
        return con

    def create_sqldb(self) -> None:
        if os.path.exists(self.sqldb_path):
            return

        with self._lock, self.sql_connect() as con:
            # Create table
            cur = con.cursor()
            cur.execute("""CREATE TABLE flows (
                        status TEXT NOT NULL,
                        formula TEXT NOT NULL,
                        workdir TEXT NOT NULL,
                        pyfile TEXT NOT NULL,
                        flow_id PRIMARY KEY,
                        upload_date TIMESTAMP,
                        end_date TIMESTAMP DEFAULT NULL,
                        user_message TEXT NOT NULL
                        );
                        """)
        con.close()

    def get_incoming_flows(self) -> list[Flow]:
        flows = []
        while True:
            try:
                in_flow = self.incoming_flow_queue.get_nowait()
                flows.append(in_flow)
            except Empty:
                break

        if flows:
            def get_record(flow):
                formula = flow[0][0].input.structure.formula
                return (str(flow.status), formula, flow.workdir, os.path.basename(flow.pyfile),
                        flow.node_id, datetime.datetime.now(), None, flow.user_message)

            with self.sql_connect() as con:
                cur = con.cursor()
                values = [get_record(flow) for flow in flows]
                cur.executemany("INSERT INTO flows VALUES (?, ?, ?, ?, ?, ?, ?, ?)", values)
            con.close()

        return flows

    def get_dataframe(self) -> pd.DataFrame:
        with self.sql_connect() as con:
            df = pd.read_sql_query("SELECT * FROM flows", con)
            #print("dtype", df["upload_date"].dtype)
            return df

    def get_json_status(self) -> dict:
        # https://stackoverflow.com/questions/25455067/pandas-dataframe-datetime-index-doesnt-survive-json-conversion-and-reconversion
        status = dict(
            dataframe=self.get_dataframe().to_json() #, date_format='iso'#date_unit='ns'),
        )
        return status

    def get_flow_and_status_by_nodeid(self, node_id):
        from abipy.flowtk.flows import Flow
        with self.sql_connect() as con:
            cur = con.cursor()
            cur.execute("SELECT workdir, status FROM flows WHERE flow_id = ?", [node_id])
            row = cur.fetchone()
        con.close()

        if row and os.path.exists(row["workdir"]):
            return Flow.from_file(row["workdir"]), row["status"]
        else:
            return None, None

    def get_sql_rows_with_node_ids(self, node_id_list):
        with self.sql_connect() as con:
            cur = con.cursor()
            query = "SELECT * FROM flows WHERE flow_id IN (%s)" % ','.join('?' * len(node_id_list))
            cur.execute(query, node_id_list)
            rows = cur.fetchall()
        con.close()
        return rows

    def groupby_status(self):
        from collections import defaultdict
        d = defaultdict(list)

        with self.sql_connect() as con:
            cur = con.cursor()
            cur.execute("SELECT * FROM flows")
            rows = cur.fetchall()

        con.close()

        for row in rows:
            d[row["status"]].append(row)

        return d

    def remove_flows_with_status(self, status):
        if status == "Running":
           raise ValueError("You cannot remove a flow that is in `Running` mode!")

        count = 0
        with self._lock, self.sql_connect() as con:
            cur = con.cursor()
            cur.execute("DELETE FROM flows WHERE status = ?", [status])
            rows = cur.fetchall()
            if rows is None:
                con.close()
                return 0

            for row in rows:
                workdir = row["workdir"]
                if not os.path.exists(workdir): continue
                os.rmdir(workdir)
                count += 1

        con.close()
        return count

    def callback(self):
        """The function that will be executed by the scheduler."""
        #locked = self._lock.acquire(blocking=False, timeout=-1)
        #if not locked: return
        #with self._lock:

        new_flows = self.get_incoming_flows()
        if new_flows:
            self.flows.extend(new_flows)

        if not self.flows: return
        excs = []

        # check status.
        for flow in self.flows:
            flow.check_status(show=False)

        # Here we just count the number of tasks in the flow that are in RUNNING or SUBMITTED status.
        # This logic clearly breaks down if there are multiple schedulers running on the same machine
        # but it's easy to implement without having to contact the resource manager
        # by calling `flow.get_njobs_in_queue()` that is quite expensive (and the sysadmin won't be happy!)
        nqjobs = 0
        for flow in self.flows:
            nqjobs += (len(list(flow.iflat_tasks(status=flow.S_RUN))) +
                       len(list(flow.iflat_tasks(status=flow.S_SUB))))

        if nqjobs >= self.max_njobs_inqueue:
            print(f"Too many jobs in the queue: {nqjobs} >= {self.max_njobs_inqueue}.\n",
                  "No job will be submitted.")
            for flow in self.flows:
                flow.check_status(show=False)
            return

        max_nlaunch = self.max_njobs_inqueue - nqjobs if self.max_nlaunches == -1 else \
                      min(self.max_njobs_inqueue - nqjobs, self.max_nlaunches)

        # This check is not perfect, we should make a list of tasks to submit
        # and then select a subset so that we don't exceeed max_ncores_used
        # Many sections of this code should be rewritten though.
        #if self.max_ncores_used is not None and flow.ncores_used > self.max_ncores_used:
        ncores_allocated = sum(flow.ncores_allocated for flow in self.flows)
        if self.max_ncores_used is not None and ncores_allocated > self.max_ncores_used:
            print("Cannot exceed max_ncores_used %s" % self.max_ncores_used,
                  ", ncores_allocated:", ncores_allocated)
            return

        # Try to restart unconverged tasks.
        for flow in self.flows:
            max_nlaunch = self.restart_unconverged(flow, max_nlaunch, excs)
            if max_nlaunch <= 0: return

        for flow in self.flows:
            self.try_to_fix_flow(flow)
            # Update the pickle file.
            flow.pickle_dump()

        #with self.handle_flow_exceptions:

        for i, flow in enumerate(self.flows):
            # Submit the tasks that are ready.
            try:
                if max_nlaunch > 0:
                    nlaunch = PyLauncher(flow).rapidfire(max_nlaunch=max_nlaunch, sleep_time=10)
                    self.nlaunch += nlaunch
                    max_nlaunch -= nlaunch
                    if nlaunch:
                        cprint("[%s] Number of launches: %d" % (time.asctime(), nlaunch), "yellow")

            except Exception as exc:
                self.register_flow_exception(i, exc)

        self.handle_flow_exception()

        for flow in self.flows:
            flow.show_status()

        #if max_nlaunch <= 0: return

        self.update_flows_and_slqdb()

    def update_flows_and_slqdb(self):

        done = []
        for i, flow in enumerate(self.flows):
            if flow.all_ok:
                flow.finalize()
                done.append(i)

        if done:
            for i in done:
                flow = self.flows[i]
                self.completed_flows.append(flow)
            self.flows = [self.flows[i] for i in range(len(self.flows)) if i not in set(done)]

        locked = []
        for i, flow in enumerate(self.flows):
            errs = self.check_deadlocks(flow)
            if errs:
                msg = "\n".join(errs)
                locked.append(i)
                flow.set_status(flow.S_ERROR, msg)

        if locked:
            for i in locked:
                flow = self.flows[i]
                self.errored_flows.append(flow)

            self.flows = [self.flows[i] for i in range(len(self.flows)) if i not in set(locked)]

        with self._lock, self.sql_connect() as con:
            now = datetime.datetime.now()
            cur = con.cursor()
            query = "UPDATE flows SET status = ?, end_date = ? WHERE flow_id = ?"

            values = [(str(flow.S_RUN), now, flow.node_id) for flow in self.flows]

            if self.completed_flows:
               values.extend([(str(flow.status), now, flow.node_id) for flow in self.completed_flows])
               self.completed_flows = []

            if self.errored_flows:
               values.extend([(str(flow.S_ERROR), now, flow.node_id) for flow in self.errored_flows])
               self.errored_flows = []

            cur.executemany(query, values)

        con.close()


def print_flowsdb_file(filepath: str) -> None:
    """
    Print flows.db file to terminal.
    """
    import sqlite3
    from abipy.tools.printing import print_dataframe
    with sqlite3.connect(filepath) as con:
        df = pd.read_sql_query("SELECT * FROM flows", con)
        #print(type(df["upload_date"]))
        print_dataframe(df, title=filepath)


def sendmail(subject: str, text: str, mailto: str,
             sender: Optional[str] = None) -> int:
    """
    Sends an e-mail with unix sendmail.

    Args:
        subject: String with the subject of the mail.
        text: String with the body of the mail.
        mailto: String or list of string with the recipients.
        sender: string with the sender address. If sender is None, username@hostname is used.

    Returns: Exit status
    """
    def user_at_host():
        from socket import gethostname
        return os.getlogin() + "@" + gethostname()

    # Body of the message.
    try:
        sender = user_at_host() if sender is None else sender
    except OSError:
        sender = 'abipyscheduler@youknowwhere'

    if is_string(mailto): mailto = [mailto]

    from email.mime.text import MIMEText
    mail = MIMEText(text)
    mail["Subject"] = subject
    mail["From"] = sender
    mail["To"] = ", ".join(mailto)

    msg = mail.as_string()

    # sendmail works much better than the python interface.
    # Note that sendmail is available only on Unix-like OS.
    from subprocess import Popen, PIPE

    _sendmail = which("sendmail")
    if _sendmail is None: return -1
    # msg is string not bytes so must use universal_newlines
    p = Popen([_sendmail, "-t"], stdin=PIPE, stderr=PIPE, universal_newlines=True)

    outdata, errdata = p.communicate(msg)
    return len(errdata)


#def _test_sendmail():
#    retcode = sendmail("sendmail_test", text="hello\nworld", mailto="nobody@nowhere.com")
#    print("Retcode", retcode)
#    assert retcode == 0



#!/usr/bin/env python
r"""
Projected phonon DOS
====================

This example shows how to plot the projected phonon DOS (PJDOS) of AlAs.
See tutorial/lesson_rf2.html
"""

#%%
# Read the Phonon DOS from the netcdf file produced by anaddb with prtdos 2
# (alternatively one can use the shell and `abiopen.py OUT_PHDOS.nc -nb`
# to open the file in a jupyter notebook.

from abipy.abilab import abiopen
import abipy.data as abidata

phdos_file = abiopen(abidata.ref_file("trf2_5.out_PHDOS.nc"))

#%%
# To plot the PJDOS with matplotlib, use:

phdos_file.plot_pjdos_type(units="cm-1", title="AlAs type-projected phonon DOS")

#%%
# For the plotly version, use:

phdos_file.plotly_pjdos_type(units="cm-1", title="AlAs type-projected phonon DOS")

#%%
# To have the projection along the cartesian directions (summed over atomic types), use

phdos_file.plot_pjdos_cartdirs_type(units="Thz", stacked=True,
        title="Type-projected ph-DOS decomposed along the three Cartesian directions.")

#%%
# To plot the PJDOS for all the inequivalent sites, use:

phdos_file.plot_pjdos_cartdirs_site(view="inequivalent", stacked=True)

#%%
# Remember to close the file
phdos_file.close()



#!/usr/bin/env python
"""Analyze the parallel efficiency of the BSE code (Haydock method with model dielectric function)"""
import sys
import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk

from itertools import product
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_inputs(paw=False):
    # Crystalline silicon
    # Calculation of the GW correction to the direct band gap in Gamma
    # Dataset 1: ground state calculation
    # Dataset 2: BSE with haydock method and model dielectric function.
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
    pseudos = abidata.pseudos("14si.pspnc") if not paw else abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")

    multi = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=2)

    ecut = 8
    multi.set_vars(
        ecut=ecut,
        pawecutdg=ecut*4 if paw else None,
        nsppol=1,
        timopt=-1,
        istwfk="*1",
        paral_kgb=0,
    )

    # This grid is the most economical, but does not contain the Gamma point.
    multi.set_kmesh(
        ngkpt=[8, 8, 8],
        shiftk=[0.0, 0.0, 0.0],
    )

    gs, bse = multi.split_datasets()

    gs.set_vars(tolvrs=1e-6,
                nband=8,
                )

    # Dataset 6 BSE equation with Model dielectric function and Haydock (only resonant + W + v)
    # Note that SCR file is not needed here
    bse.set_vars(
        optdriver=99,
        ecutwfn=ecut,
        ecuteps=4.0,
        inclvkb=2,
        bs_algorithm=2,        # Haydock
        bs_haydock_niter=60,   # No. of iterations for Haydock
        bs_exchange_term=1,
        bs_coulomb_term=21,    # Use model W and full W_GG.
        mdf_epsinf=12.0,
        bs_calctype=1,         # Use KS energies and orbitals to construct L0
        mbpt_sciss="0.8 eV",
        bs_coupling=0,
        bs_loband=2,
        nband=8,
        bs_freq_mesh="0 10 0.1 eV",
        bs_hayd_term=0,        # No terminator
        #gwmem=01               # Compute the model-dielectric function on-the-fly.
    )

    return gs, bse


def build_flow(options):
    """
    Build an `AbinitWorkflow` used for benchmarking ABINIT.
    """
    gs_inp, bse_inp = make_inputs(paw=options.paw)
    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)

    gs_work = flowtk.Work()
    gs_work.register_scf_task(gs_inp)
    flow.register_work(gs_work)
    flow.exclude_from_benchmark(gs_work)

    mpi_list = options.mpi_list

    if options.mpi_list is None:
        nkpt = len(gs_inp.abiget_ibz().points)
        ntrans = (2*2*nkpt)**2
        mpi_list = [p for p in range(1, 1 + ntrans) if ntrans % p == 0]
    if options.verbose: print("Using mpi_list:", mpi_list)

    bse_work = flowtk.Work()
    for mpi_procs, omp_threads in product(mpi_list, options.omp_list):
        if not options.accept_mpi_omp(mpi_procs, omp_threads): continue
        manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
        bse_work.register_bse_task(bse_inp, manager=manager, deps={gs_work[0]: "WFK"})
    flow.register_work(bse_work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return

    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
"""
This module provides functions to compute integrals of Bessel functions.
"""
from __future__ import annotations

import numpy as np

from scipy.special import spherical_jn
from scipy.interpolate import UnivariateSpline
from scipy.integrate import simps # cumtrapz, quad


_DEFAULTS = {"numq": 3001, "numr": 3001}


def spline_int_jlqr(l, qmax, rcut, numq=None, numr=None):
    r"""
    Compute :math:`j_n(z) = \int_0^{rcut} r^2 j_l(qr) dr`
    where :math:`j_l` is the Spherical Bessel function.

    Args:
        l: Angular momentum
        qmax: Max :math:`|q|` in integral in Ang-1
        rcut: Sphere radius in Angstrom.
        numq: Number of q-points in qmesh.
        numr: Number of r-points for integration.

    Return:
        Spline object.
    """
    numq = numq if numq is not None else _DEFAULTS["numq"]
    numr = numr if numr is not None else _DEFAULTS["numr"]

    rs = np.linspace(0, rcut, num=numr)
    r2 = rs ** 2
    qmesh = np.linspace(0, qmax, num=numq)
    values = []
    for q in qmesh:
        qrs = q * rs
        #qrs = 2 * np.pi * q * rs
        ys = spherical_jn(l, qrs) * r2
        intg = simps(ys, x=rs)
        values.append(intg)

    return UnivariateSpline(qmesh, values, s=0)



#!/usr/bin/env python
r"""
Phonopy with AbiPy flow
=======================

This example shows how to compute phonon frequencies with phonopy
(supercells and finite-difference method).

This approach could be useful to obtain vibrational properties with XC functionals
for which DFPT is not yet implemented.

.. warning:

    This example requires the `phonopy package <http://atztogo.github.io/phonopy/examples.html>`_
"""

import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk

from abipy.flowtk.abiphonopy import PhonopyWork


def build_flow(options):
    """
    Create a `Flow` for phonon calculations with phonopy:
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Initialize structure and pseudos
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
    pseudos = abidata.pseudos("14si.pspnc")

    # Build input for GS calculation.
    gsinp = abilab.AbinitInput(structure, pseudos)
    gsinp.set_vars(ecut=4, nband=4, toldff=1.e-6)

    # This gives ngkpt = 4x4x4 with 4 shifts for the initial unit cell.
    # The k-point sampling will be rescaled when we build the supercell in PhonopyWork.
    gsinp.set_autokmesh(nksmall=4)
    #gsinp.set_vars(ngkpt=[4, 4, 4])

    flow = flowtk.Flow(workdir=options.workdir)

    # Use a 2x2x2 supercell to compute phonons with phonopy
    work = PhonopyWork.from_gs_input(gsinp, scdims=[2, 2, 2])
    flow.register_work(work)

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())

############################################################################
#
# Run the script with:
#
#     run_phonopy_si.py -s
#
# the output results are produced in ``flow_phonopy_si/w0/outdata/``
# Follow the instructions in the README file:
#
# .. code-block:: md
#
#    To plot bands, use:
#            phonopy -p band.conf
#
#    To plot phonon dos, use:
#            phonopy -p dos.conf
#
#    To plot bands and dos, use:
#            phonopy -p band-dos.conf
#
#    See also:
#            http://atztogo.github.io/phonopy/examples.html
#            http://atztogo.github.io/phonopy/setting-tags.html#setting-tags
#
# The command:
#
#   phonopy -p band-dos.conf
#
# will produce:
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_phonopy_si.png?raw=true
#    :alt: Phonon Band structure computed with phonopy.
#
